import numpy as np
def model_Inducible_Double_Uptake_LogisticGrowth(y, t, params):
	x    = y[0]
	inde = y[1]
	indi = y[2]
	m    = y[3]
	p    = y[4]

	mu_max = params[0]
	x_max  = params[1]
	upind  = params[2]
	k_ind  = params[3]
	synm   = params[4]
	degm   = params[5]
	synp   = params[6]
	n_ind  = params[7]

	mu = mu_max*(1 - x/x_max)
	
	dx = mu*x
	dinde= -upind*inde
	dindi= upind*inde
	dm = synm*indi**n_ind/(indi**n_ind + k_ind**n_ind) - degm*m
	dp = synp*m - mu*p

	return np.array([dx, dinde, dindi, dm, dp])

x,inde,indi,m,p,mu_max,x_max,upind,k_ind,synm,degm,synp,n_ind= np.random.rand(13)*10

x,inde,indi,m,p,mu_max,x_max,upind,k_ind,synm,degm,synp,n_ind= list(map(float, [x,inde,indi,m,p,mu_max,x_max,upind,k_ind,synm,degm,synp,n_ind]))

y = [x,inde,indi,m,p]

t = 0
dt = 1e-3

params = mu_max,x_max,upind,k_ind,synm,degm,synp,n_ind

y = y + dt*model_Inducible_Double_Uptake_LogisticGrowth(y, t, params)

y = y + dt*model_Inducible_Double_Uptake_LogisticGrowth(y, t, params)


from os       import getcwd, listdir
from os.path  import abspath, dirname, join

__all__ = ['models',
           'aicanalysis',
           'curvefitting',
           'sensitivityanalysis',
           'simulation',
           'strike_goldd_simplified',
           'traceanalysis'
           ]

#Add styles
library = join(dirname(__file__), 'stylelib')
styles  = {file.split('.')[0]: abspath(join(library,file)) for file in listdir(library)}


import numpy  as np
import pandas as pd

###############################################################################
#Template Based Checkers
###############################################################################
def check_and_assign_param_values(core_model, parameters):
    '''
    Checks if parameters are valid and returns a DataFrame.
    
    Accepts parameters as:
        1. dict of floats/np.float where keys are parameter names
        2. dict of arrays  where keys are parameter names
        3. Series where where indices are parameter names
        4. DataFrame where columns are parameter names
    
    Parameter names may be suffixed with an underscore i.e. param1_1_1 will
    be treated as param1_1
    
    Raises an error if parameter names cannot be matched.
    
    Returns a DataFrame of parameters with columns in order.
    '''
    if type(parameters) == pd.DataFrame:
        parameter_df = pd.DataFrame(parameters)
    elif type(parameters) == pd.Series:
        parameter_df = pd.DataFrame(parameters).T
    elif type(parameters) == dict:
        parameter_df = pd.DataFrame(parameters)
        try:
            parameter_df = pd.DataFrame(parameters)
        except:
            parameter_df = pd.DataFrame([parameters])
    else:
        raise Exception('Error in parameters. Parameters must be dict, DataFrame or Series.')
    
    parameter_df.reset_index(drop=True, inplace=True)
    
    df_columns        = list(parameter_df.columns)
    core_model_params = core_model['parameters'] + core_model['inputs']
    
    #Check that parameter names match
    if set(df_columns).difference(core_model_params):#Not match
        #Assume parameters have been suffixed
        #Remove the suffix, reassign and rearrange
        df_columns_ = ['_'.join(p.split('_')[:-1]) for p in df_columns]
        
        if set(df_columns_).difference(core_model_params):
            raise Exception('Could not match parameter names with that of core_model')
        else:
            parameter_df.columns = df_columns_
            
            parameter_df         = parameter_df[core_model_params]
            return parameter_df
    else:#Match
        #Ensure order is correct
        parameter_df = parameter_df[core_model_params]
        return parameter_df

def check_and_assign_solver_args(solver_args):
    '''
    Accepts a dict of solver_args and checks if it is valid against a template.
    
    Raises an exception if 
        1. Unexpected keys are found.
        2. Value indexed by tcrit cannot be converted into a list
        3. Value indexed by other keys cannot be converted in a float
    
    Returns the template but with new values from solver_args.
    '''
    
    template = {'rtol'   : 1.49012e-8,
                'atol'   : 1.49012e-8,
                'tcrit'  : [],
                'h0'     : 0.0,
                'hmax'   : 0.0,
                'hmin'   : 0.0,
                'mxstep' : 0
                }
    
    if not solver_args:
        return template
    
    for key in solver_args:
        if key not in template:
            raise Exception('Error in solver_args. Unexpected key found: ' + str(key))
        
        elif key == 'tcrit':
            try:
                template[key] = list(solver_args[key])
            except:
                raise Exception('Error in solver_args. value indexed by key ' + str(key) + ' must be list-like.')
        else:
            try:
                template[key] = float(solver_args[key])
            except:
                raise Exception('Error in solver_args. value indexed by key ' + str(key) + ' must be a number.')                                

    return template

def check_and_assign_init(states, init, init_orient='scenario'):
    '''
    Accepts init in the form of {scenario_num : array} if init_orient is "scenario".
    Accepts init in the form of {state : array} if init_orient is "state".
    
    Formats init in the form required by the models datastructure.
    
    Accepts init in the form:
        1. {scenario_num : array} if init_orient is scenario
        2. {state : array} if init_orient is "state"
        3. A DataFrame where the columns are the states if init_orient is "state"
        4. A Series where the indices are the states if init_orient is "state"
        5. A numpy array where each column corresponds the states in order
    
    Raises an exception in (1) if the length of any array does not match the length
    of states.
    
    Raises an exception in (2) if 
    
    '''
    def check_valid_scenario_num(scenario_num):
        if type(scenario_num) != int:
            raise Exception('Error in init argument. Only positive integers can be used as scenario_num.')
        elif scenario_num < 1:
            raise Exception('Error in init argument. Only positive integers can be used as scenario_num.')
                
        else:
            return
    
    def check_valid_num_states(states, row):
        if len(states) == len(row):
            return
        else:
            raise Exception('Error in init. Number of states does not match core model. Expected:\n' + str(states) + '\nDetected:\n' + str(row))
    
    def check_valid_states(states, columns):
        if set(states).difference(columns):
            raise Exception('Error in init. Unexpected states found. Expected:\n' + str(states) + '\nDetected:\n' + str(columns))
    if init is not None:
        if init_orient == 'scenario':
            for key in init:
                check_valid_num_states(states, init[key])
                check_valid_scenario_num(key)
            return init
        
        else:
            if type(init) == dict:
                init1 = pd.DataFrame([init])
                init1.index += 1
                check_valid_states(states, list(init1.columns))
            elif type(init) == pd.DataFrame:
                init1 = init.reset_index(drop=True)
                init1.index += 1
                check_valid_states(states, list(init1.columns))
            elif type(init) == pd.Series:
                init1 = pd.DataFrame(init).T
                init1.index += 1
                check_valid_states(states, list(init1.columns))
            
            else: #Array based
                init1 = pd.DataFrame(init)
                init1.index += 1
                try:
                    init1.columns = states
                except:
                    raise Exception('Number of states does not match width of init')
            
            check_valid_num_states(states, list(init1.columns))
            [check_valid_scenario_num(x)  for x in list(init1.index)]
                    
            return init1
    else:
        return {1: np.array([0]*len(states))}

def check_and_assign_parameter_bounds(parameter_df, parameter_bounds):
    '''
    Accepts parameters as DataFrame. Parameter names must match core model.
    '''
    bounds1 = {}
    for key in parameter_df:
        if key in parameter_bounds:
            min_val, max_val = parameter_bounds[key]
        else:
            min_val = float(np.min(parameter_df[key]))/10
            max_val = float(np.max(parameter_df[key]))*10
            
        bounds1[key] = [min_val, max_val]
    return bounds1

def check_and_assign_priors(parameter_df, priors):
    '''
    Parameter names must match core model.
    '''
    priors1 = {}
    if priors is None:
        return priors1
    
    for key in priors:
        if key not in parameter_df:
            raise Exception('Unexpected parameter name in priors: ' + str(key))
        try:
            priors1[key] = np.array(priors[key])
        except:
            raise Exception('Error in ' + str(key) + '. Values indexed by priors must be list-like.')
    
    return priors1

def check_and_assign_tspan(tspan):
    if tspan:
        try:
            tspan1 = [np.linspace(*span) for span in tspan]
        except:      
            try:
                tspan1 = [np.array(span) for span in tspan]
            except:
                raise Exception('Error in tspan. tspan must be a list of list-like structures.')
            
    else:
        tspan1 = [np.linspace(0, 600, 31)]
    return tspan1


import numpy  as np
import pandas as pd
###############################################################################
#Check if Analysis Settings
###############################################################################    
def is_analysis_settings(config, section):
    if section in ['id', 'system_type', 'states', 'parameters', 'inputs', 'equations', 'descriptions', 'ia']:
        return False
    if 'system_type' in config[section]:
        return False
    
    return True

###############################################################################
#Supporting Functions for Reading Strings into Python
###############################################################################    
def eval_init_string(string):
    try:
        result = string_to_array(string)
        result = {i+1: result[i] for i in range(len(result))}
    except:
        result        = string_to_dataframe(string)
        result.index += 1
        result        = result.T.to_dict('list')

    for key in result:
        if any([type(x) == str for x in result[key]]) == str:
            raise Exception('Error in reading initial values. There is likely a missing comma(s).')
        try:
            result[key] = np.array(result[key])
        except:
            raise Exception('Error in reading initial values. Ensure you input is list-like.')
    return result

def eval_tspan_string(string):
    if string:
        try:
            return string_to_linspace(string)
        except:
            return np.array(eval(string.replace('\n', '')))
    else:
        return [np.linspace(0, 600, 31)]

def eval_params_string(string):
    return string_to_dataframe(string)

def string_to_linspace(string):
    '''
    Use this when you expect arguments for numpy.linspace
    '''
    try:
        return [np.linspace(*segment) for segment in eval('[' + string +']')]
    except:
        try:
            return [np.linspace(*segment) for segment in eval(string)]
        except:
            return np.array(eval(string))

def string_to_list(string):
    '''
    Use this when you expect a list of numbers
    '''
    string1 = string.replace('\n', ',')
    lst     = [s.strip() for s in string1.split(',')]
    lst     = [s for s in lst if s]
    string1 = ', '.join(lst)

    try:
        return eval('[' + string1 +']')
    except:
        return eval(string1)

def string_to_list_string(string):
    '''
    Use this when you expect a list of strings
    '''
    string1 = string.replace('[', '').replace(']', '').replace('\n', ',')
    lst     = [s.strip() for s in string1.split(',')]
    lst     = [s for s in lst if s]

    return lst
    
def string_to_array(string):
    '''
    Use this when you expect a list that can be converted into a numpy array.
    '''
    return np.array(string_to_list(string))
        
def string_to_dict(string):
    '''
    Use this when you expect a dictionary
    '''
    result = [s.strip() for s in split_at_top_level(string)]
    result = [line.split('=') for line in result ]
    result = [[lst[0].strip(), '='.join(lst[1:]).strip()] for lst in result]
    
    result = {pair[0]: try_eval(pair[1]) for pair in result}    
    
    return result

def string_to_dict_array(string):
    '''
    Use this when you expect a dictionary and expect the values to be numpy arrays
    '''
    temp = string_to_dict(string)
    try:
        iter(temp[next(iter(temp))])
        return {key: np.array(temp[key])  for key in temp}
    except:
        return {key: np.array([temp[key]])  for key in temp}
    
def string_to_dataframe(string):
    '''
    Use this when you expect a dictionary and want to convert it into a DataFrame
    '''
    return pd.DataFrame(string_to_dict_array(string))
    
def split_at_top_level(string, delimiter=','):
    '''
    Use this for nested lists.
    This is also a helper function for string_to_dict.
    '''
    nested = []
    buffer = ''
    result = []
    matching_bracket = {'(':')', '[':']', '{':'}'}
    
    for char in string:
        if char in ['[', '(', '{']:
            nested.append(char)
            buffer += char
        
        elif char in [']', ')', '}']:
            if char == matching_bracket.get(nested[-1]):
                nested = nested[:-1]
                buffer += char
            else:
                raise Exception('Mismatched brackets.' )
        elif char == delimiter and not nested:
            if buffer:
                result.append(buffer)
                buffer = ''
        else:
            buffer += char
    if buffer:
        result.append(buffer)
    return result

def try_eval(x):
    '''
    Attempts to convert string to numbers.
    '''
    try:
        return eval(x)
    except:
        return x


import matplotlib.pyplot as plt 
import numpy             as np
import os
import pandas            as pd
import re
import seaborn           as sns
from matplotlib          import get_backend
from numba               import jit
from scipy.integrate     import odeint

from numba import jit

###############################################################################
#Globals
###############################################################################
#Refer for details: https://seaborn.pydata.org/tutorial/color_palettes.html
palette_types = {'color':     lambda n_colors, **kwargs : sns.color_palette(n_colors=n_colors,     **{**{'palette': 'muted'}, **kwargs}),
                 'light':     lambda n_colors, **kwargs : sns.light_palette(n_colors=n_colors+2,   **{**{'color':'steel'}, **kwargs})[2:],
                 'dark' :     lambda n_colors, **kwargs : sns.dark_palette( n_colors=n_colors+2,   **{**{'color':'steel'}, **kwargs})[2:],
                 'diverging': lambda n_colors, **kwargs : sns.diverging_palette(n=n_colors,        **{**{'h_pos': 250, 'h_neg':15}, **kwargs}),
                 'cubehelix': lambda n_colors, **kwargs : sns.cubehelix_palette(n_colors=n_colors, **kwargs),
                 }    
#Refer for details: https://xkcd.com/color/rgb/
all_colors    = sns.colors.xkcd_rgb

###############################################################################
#Integration
###############################################################################
def piecewise_integrate(function, init, tspan, params, model_num, scenario_num, modify_init=None, modify_params=None, solver_args={}, solver=odeint, overlap=True, args=()):
    '''Piecewise integration function with scipy.integrate.odeint as default. 
    Can be changed using the solver argument.

    Parameters
    ----------
    function : function
        A function for numerical integration that takes in the form f(states, 
        time, params, args). states, params and args are numpy arrays while time 
        is a float or numpy float. Will be integrated using scipy.integrate.odeint
    init : numpy.array
        Initial values of states for numerical integration.
    tspan : list of numpy.array
        Segments for piecewise integration where each segment is a numpy array of 
        time points in ascending order.
    params : numpy.array
        An array of parameter values.
    model_num : int
        Number of the model.
    scenario_num : int
        Number of the scenario.
    modify_init : function, optional
        Function for modifying initial values before integrating over a time 
        segment. The default is None.
    modify_params : function, optional
        Function for modifying parameter values before integrating over a time 
        segment. The default is None.
    solver_args : dict, optional
        Dictionary of keyword arguments for scipy.integrate.odeint. The default is {}.
    solver : scipy.integrate.odeint, optional
        The integrating function. The default is odeint. Do not touch unless you 
        know what you are doing.
    overlap : bool, optional
        Avoids double counting between segments. In general, False is required 
        only for plotting. The default is True.
    args : tuple, optional
        A tuple of arguments. The default is ().

    Returns
    -------
    y_model : numpy.array
        An array of values from numerical integration.
    t_model : numpy.array
        An array of time points.

    '''
    solver_args1 = solver_args if solver_args else {}
    
    tspan_     = tspan[0]
    init_      = modify_init(init_values=init,    params=params, model_num=model_num, scenario_num=scenario_num, segment=0) if modify_init   else init
    params_    = modify_params(init_values=init_, params=params, model_num=model_num, scenario_num=scenario_num, segment=0) if modify_params else params 
    y_model    = solver(function, init_, tspan_, args=tuple([params_]) + args, **solver_args1)
    t_model    = tspan[0]
    
    for segment in range(1, len(tspan)):
        tspan_   = tspan[segment]
        init_    = modify_init(init_values=y_model[-1], params=params, model_num=model_num, scenario_num=scenario_num, segment=segment) if modify_init   else y_model[-1]
        params_  = modify_params(init_values=init_,     params=params, model_num=model_num, scenario_num=scenario_num, segment=segment) if modify_params else params 
        y_model_ = solver(function, init_, tspan_, args=tuple([params_]) + args, **solver_args1)       
        y_model  = np.concatenate((y_model, y_model_), axis=0) if overlap else np.concatenate((y_model[:-1],  y_model_), axis=0)   
        t_model  = np.concatenate((t_model, tspan_), axis=0)   if overlap else np.concatenate((t_model[:,-1], tspan_),   axis=0)

    return y_model, t_model

#Templates for modify_init
def modify_init(init_values, params, model_num, scenario_num, segment):
    '''
    Return a new np.array of initial values. For safety, DO NOT MODIFY IN PLACE.
    
    :meta private:
    '''
    new_init = init_values.copy()
    return new_init

def modify_params(init_values, params, model_num, scenario_num, segment):
    '''
    Return a new np.array of initial values. For safety, DO NOT MODIFY IN PLACE.
    
    :meta private:
    '''
    new_params = params.copy()
    return new_params

###############################################################################
#Multi-Model Integration
###############################################################################
def integrate_models(models, params, *extra_variables, args=(), mode='np', overlap=True):
    '''Integrates models with params    

    Parameters
    ----------
    models : dict
        A dictionary of model data structures.
    params : pandas.DataFrame
        A DataFrame of parameter values for integration.
    *extra_variables : function
        Additional functions for evaluating the integrated results.
    args : tuple, optional
        Additional arguments for the functions to be integrated. The default is ().
    mode : {'np', 'df'}, optional
        Use 'np' if the functions in extra_variables are meant to work with numpy 
        arrays and 'pd' if functions are meant to work with DataFrames. The 
        default is 'np'.
    overlap : TYPE, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    dict
        A dictionary of the integrated results.
    dict
        A dictionary of the calculated extra variables.
    '''
    print('Simulating models')
    y_models = {}
    e_models = {v: {model_num: {scenario_num: {} for scenario_num in models[model_num]['init']} for model_num in models} for v in extra_variables}
    
    
    if type(params) == dict:
        try: 
            params_ = pd.DataFrame(params)
        except:
            params_ = pd.DataFrame([params])

        return integrate_models(models, params_, *extra_variables, args=args, overlap=overlap)
    
    if len(params.shape) == 1:
        return integrate_models(models, np.array([params]), *extra_variables, args=args, overlap=overlap)
    
    else:
        for model_num in models:
            model               = models[model_num]
            param_names         = model['params']
            params1             = params[param_names]
            y_models[model_num] = {}
            
            for scenario_num in models[model_num]['init']:
                if type(scenario_num) == int:
                    if scenario_num < 1:
                        continue
                y_models[model_num][scenario_num] = {}
                
                for name, row in params1.iterrows():
                    y_model, t_model = piecewise_integrate(params        = row.values,
                                                           function      = model['function'], 
                                                           init          = model['init'][scenario_num],
                                                           tspan         = model['tspan'],
                                                           modify_init   = model['int_args']['modify_init'],
                                                           modify_params = model['int_args']['modify_params'],
                                                           solver_args   = model['int_args']['solver_args'],
                                                           model_num     = model_num, 
                                                           scenario_num  = scenario_num, 
                                                           overlap       = overlap,
                                                           args          = args
                                                           )
                    y_models[model_num][scenario_num][name] = pd.DataFrame(y_model, columns=models[model_num]['states'])
                    
                    
                    for func in extra_variables:
                        y_model_ = y_models[model_num][scenario_num][name] if mode == 'pd' else y_model
                        variable = func(y_model_, t_model, row.values)
                        e_models[func][model_num][scenario_num][name] = variable 
                                       
            y_models[model_num][0] = t_model
        
        print('Simulation complete')
        return y_models, e_models

###############################################################################
#Plot
###############################################################################
def plot_model(plot_index, y, e={}, titles={}, labels={}, figs=(), AX={}, palette={}, line_args={}, legend_args={}):
    '''Plots the integrated results.

    Parameters
    ----------
    plot_index : dict
        A dictionary of states and extra variables to plot.
    y : dict
        The values of the integrated results from integrate_models.
    e : dict, optional
        The values of the integrated results from integrate_models. The default 
        is {}.
    titles : dict, optional
        A dictionary of axes titles. The default is {}.
    labels : TYPE, optional
        A dictionary of line labels. The default is {}.
    figs : list of matplotlib.Figure, optional
        A list of Figure objects that will be maximized after plotting. The 
        default is ().
    AX : dict of matplotlib.Axes, optional
        A dictionary of Axes objects for plotting. If {}, the Axes will be 
        automatically generated. The default is {}.
    palette : dict, optional
        A dictionary that controls the color of the plots. The default is {}.
    line_args : dict, optional
        Keyword arguments for the plot method of the Axes class. The default is 
        {}.
    legend_args : dict, optional
        Keyword arguments for controlling the appearance of the legend. The 
        default is {}.

    Returns
    -------
    figs : list
        A list of Figure objects.
    AX : dict
        A dict of Axes objects.

    '''
    figs1, AX1 = (figs, AX) if AX else make_AX(plot_index)
    palette1   = palette if palette else {'palette_type': 'color'}
    colors     = make_colors_from_simulation(y, e, palette1)

    for model_num in plot_index:
        for state in plot_index[model_num]:
            try:
                ax = AX1[model_num][state]
            except:
                raise Exception('No Axes provided for state ' + str(state) + ' in model ' + str(model_num) +'.')
            
            if callable(state):
                temp = e[state]
            else:       
                temp = y
                
            for scenario_num in temp[model_num]:
                if scenario_num < 1:
                    continue
                
                first=True
                for row_name in temp[model_num][scenario_num]:
                    color = colors[model_num][scenario_num]
                    if type(color) == dict:
                        color = color[row_name]                        
                    x_arr = e[state][model_num][scenario_num][row_name][1] if callable(state) else y[model_num][0] 
                    y_arr = e[state][model_num][scenario_num][row_name][0] if callable(state) else y[model_num][scenario_num][row_name][state]
                    label = get_label(labels, model_num, scenario_num, first=first, row_name=row_name)
                    
                    ax.plot(x_arr, y_arr, label=label, color=color, **line_args)
                    ax.ticklabel_format(style='sci', scilimits=(-2,3))
                    
                    first = False

    [fs(fig) for fig in figs1]
    apply_titles_and_legend(AX1, titles, legend_args)
                
    return figs1, AX1 

###############################################################################
#Supporting Functions. Do not run.
############################################################################### 
def get_label(labels, model_num, scenario_num, first, row_name=None):
    '''
    :meta private:
    '''
    if row_name is None:
        label = labels.get(model_num, {}).get(scenario_num, '')
    else:
        label = labels.get(model_num, {}).get(scenario_num, {})
        
        if type(label) == str:
            if not first:
                label = None
        else:
            try:
                label = label[row_name]
            except:
                if not first:
                    label = None
        
    return label

def apply_titles_and_legend(AX, titles, legend_args):
    '''
    :meta private:
    '''
    for model_num in AX:
        for state in AX[model_num]:
            ax    = AX[model_num][state]
            title = titles.get(model_num, {}).get(state, '')
            if title:
                ax.set_title(title)
            if legend_args:
                ax.legend(**legend_args)    
    
def fs(figure):
    '''
    :meta private:
    '''
    try:
        plt.figure(figure.number)
        backend   = get_backend()
        manager   = plt.get_current_fig_manager()
        
        if backend == 'TkAgg':
            manager.resize(*manager.window.maxsize())
        
        elif backend == 'Qt5Agg' or backend == 'Qt4Agg': 
            manager.window.showMaximized()
        
        else:
            manager.frame.Maximize(True)
        plt.pause(0.03)
    except:
        pass
    return figure

###############################################################################
#Color Generation
###############################################################################    
def make_colors_from_simulation(y, e={}, palette={'palette_type': 'color'}):
    '''
    :meta private:
    '''
    global palette_types
    global all_colors
    
    if type(palette[next(iter(palette))]) == dict: #Assume full dict in this case
        return palette
    
    scenario_names = set()
    tree           = {}
    def counter(y):
        for model_num in y:
            if model_num not in tree:
                tree[model_num] = {}
            for scenario_num in y[model_num]:
                if scenario_num > 0:
                    scenario_names.add(scenario_num)
                    if scenario_num not in tree[model_num]:
                        tree[model_num][scenario_num] = set()
                    tree[model_num][scenario_num].update([name for name in list(y[model_num][scenario_num].keys())]) 
                    
    
    counter(y), [counter(e[func]) for func in e]
    
    n_scenarios = len(scenario_names)
    
    func   = palette_types[palette['palette_type']]
    kwargs = {key: palette[key] for key in palette if key != 'secondary' and key != 'palette_type'}
    
    if 'color' in kwargs:
        try:
            kwargs['color'] = all_colors[kwargs['color']]
        except:
            pass
        
    base_colors = dict(zip(scenario_names, func(n_scenarios, **kwargs)))
    
    colors = {}
    if 'secondary' in palette:
        func   = palette_types[palette['secondary']]
        colors = {m: {s: dict(zip(tree[m][s], func(len(tree[m][s]), color=base_colors[s])))  for s in tree[m]} for m in tree}        
    else:
        colors = {m: {s: {r : base_colors[s] for r in tree[m][s]}  for s in tree[m]} for m in tree}        
        
    return colors
         
 
###############################################################################
#Axes Generation
###############################################################################
def make_AX(plot_index={}, data={}):
    '''
    Generates a dictionary of axes objects using either plot_index OR data.
    
    :meta: private
    '''
    dataset_keys = plot_index if plot_index else {model_num: list(data[model_num].keys()) for model_num in data}
    max_len      = 0

    for model_num in dataset_keys:
        max_len = max(max_len, len(plot_index[model_num]))
    
    figs = [plt.figure() for i in range(max_len)]
    AX_  = {}
    AX   = {}
    for i in range(max_len):
        fig = figs[i]
        cols = 0
        for model_num in dataset_keys:
            try:
                key           = dataset_keys[model_num][i]
                cols += 1
            except:
                pass
        AX_[i] = [fig.add_subplot(1, cols, i+1) for i in range(cols)]
        
        c = 0
        for model_num in dataset_keys:
            try:
                key           = dataset_keys[model_num][i]
                
                if model_num not in AX:
                    AX[model_num] = {}
                    
                AX[model_num][key] = AX_[i][c]
 
                c += 1
            except:
                pass
    return figs, AX

def export_simulation_results(y, e, prefix='', directory=None):
    '''Save the simulation results in csv format.

    Parameters
    ----------
    y : dict
        The simulation results returned by integrate_models.
    e : dict
        The simulation results for the extra variabes returned by integrate_models.
    prefix : str, optional
        A prefix for each file generated. The default is ''.
    directory : TYPE, optional
        The directory for storing the exported files. The default is None.

    Returns
    -------
    results : list
        Names of files generated.

    '''
    results = []
    
    if directory:
        cwd = os.getcwd()
        try:
            os.mkdir(directory)
        except:
            pass
        os.chdir(directory)
        
    for model_num in y:
        time = y[model_num][0]
        for scenario_num in y[model_num]:
            if scenario_num < 1:
                continue
            for row_name in y[model_num][scenario_num]:
                filename = '_'.join(['y', str(model_num), str(scenario_num), str(row_name)]) + '.csv'
                filename = prefix + '_' + filename if prefix else filename
                table    = y[model_num][scenario_num][row_name]
                table.insert(0, 'Time', value=time)
                table.to_csv(filename)
                
                results.append(filename)
    
    for func in e:
        doc     = func.__doc__
        columns = search_string_for_x_y(doc)
        columns = columns['x'], columns['y']
        for model_num in e[func]:
            for scenario_num in e[func][model_num]:
                if scenario_num < 1:
                    continue
                for row_name in e[func][model_num][scenario_num]:
                    y, x, _  = e[func][model_num][scenario_num][row_name]
                    table    = pd.DataFrame(np.stack((x,y), axis=1), columns=['x', 'y'])
                    filename = '_'.join([func.__name__, str(model_num), str(scenario_num), str(row_name)]) + '.csv'
                    filename = prefix + '_' + filename if prefix else filename
                    table.to_csv(filename)
                    
                    results.append(filename)
    
    if directory:
        os.chdir(cwd)

    return results

def search_string_for_x_y(string):
    '''
    :meta private:
    '''
    result = {'x' :'x',
              'y' :'y'
              }
    if not string:
        return result
    
    for key in ['x', 'y']:
        
        match = re.search('(?<=' + key + '=).*', string)
        if not match:
            pattern = re.search(key + '(\s+)*=', string)
            if pattern:
                pattern = '(?<=' + pattern + ').*'
                match   = re.search(pattern, string)

        if match:
            result[key] = match[0].strip()
        
    return result  
  
if __name__ == '__main__':
    pass


# -*- coding: utf-8 -*-
"""
This module contains functions to generate SBML standard files.

Features:
-to generate SBML .xml files from a list of configuration files
-to generate SBML files from model database
-to autogenerate SBML .xml files for all configuration files contained inside a folder path  
"""

import os
import glob

import BMSS.standardfiles_generators.simplesbml as simplesbml
import BMSS.models.model_handler    as mh
import BMSS.models.settings_handler as sh

def unitlookup(settings):
    '''Reads in the list of units listed in the database and converts them to 
    units defined in simplesbml
    
    :meta private:
    '''
    
    unit_model = settings['units']
    for w in unit_model: #Unit Look-up/conversion
        unit_model[w] = unit_model[w].replace('aa', 'items') 
        unit_model[w] = unit_model[w].replace('mol-1', 'permol')
        unit_model[w] = unit_model[w].replace('molmin-1', 'molpermin')
        unit_model[w] = unit_model[w].replace('1/', 'per_') 
        unit_model[w] = unit_model[w].replace('M/', 'molper') 
        unit_model[w] = unit_model[w].replace('NONE', 'Dimension_less') 
        unit_model[w] = unit_model[w].replace('dimensionless', 'Dimension_less')
        unit_model[w] = unit_model[w].replace('/l', 'L') 
        unit_model[w] = unit_model[w].replace('sec', 'second')
        unit_model[w] = unit_model[w].replace('molL-1min-1', 'molperLmin') 
        unit_model[w] = unit_model[w].replace('min-1', 'per_min')
        unit_model[w] = unit_model[w].replace('molL-1', 'molperL')   
        
    return unit_model

def SBMLcreation(core_model, settings, unit_model, addparam, init_scenario, param_scenario):
    '''Reads in the core model, settings, parameters and which scenario of 
    init and parameters values and outputs SBML format in sbmlstr
    
    :meta private:
    '''
    
    model_sbml = simplesbml.SbmlModel()
                
    for i in range(len(core_model['states'])):
        model_sbml.addSpecies(core_model['states'][i], settings['init'][init_scenario+1][i])    
    
    for i in range(len(addparam.columns)):
        name = addparam.columns[i]
        model_sbml.addParameter(name, addparam.values[param_scenario,i], units = unit_model[name])
    
    for eqn in core_model['equations']:
        if eqn is not '':
            if eqn[0] == 'd': #if equation is for deritive with time
                variable_start = eqn.index("d")
                variable_end = eqn.index("=")
                eqn = eqn.replace('**', '^')  
                variable_name = eqn[variable_start+1:variable_end-1]
                model_sbml.addRateRule(variable_name, eqn[variable_end+2:])
                
            else: #if equation is for variable directly
                variable_end = eqn.index("=")
                assignment_name = eqn[:variable_end-1]
                eqn = eqn.replace('**', '^') 
                model_sbml.addAssignmentRule(assignment_name, eqn[variable_end+2:])
                
    sbmlstr = model_sbml.toSBML()
    return sbmlstr

  
def config_to_sbml(inifileslist):
    '''Read in list of configuration files and generate the corresponding SBML
    files.
    
    :param inifileslist: a list of input configuration files in strings.
    :type inifileslist: list
    '''
    
    sbmlfilelist=[]
    number_scenario = 0
    
    for f in inifileslist:
        print('\n', f)
    
       
    for f in inifileslist:    #Add ini files to database
        print('filename', f)
        system_type = mh.config_to_database(f)
        system_types_settings_names = sh.config_to_database(f)
        system_type, settings_name = system_types_settings_names[0]
        search_result_model = mh.quick_search(system_type)
        search_result_settings = sh.quick_search(system_type=system_type, settings_name=settings_name)
        core_model    = search_result_model
        settings = search_result_settings
        addparam =  settings['parameters']
        number_init = len(settings['init'])
        number_parameters = len(settings['parameters'])
        unit_model = unitlookup(settings)
    
        for j in range(number_init): #Cycle through number of init values
            for k in range(number_parameters): #Cycle through number of parameters
                sbmlstr = SBMLcreation(core_model, settings, unit_model, addparam, j, k)
                #Function that creates SBML file and returns the number of files outputed and the file list.
                number_scenario = number_scenario + 1
                placeholder = os.path.splitext(f)[0]
                sbmlfilename = 'DatabasetoSBML_' + placeholder +'.xml'
                f = open(sbmlfilename, 'w') #creates SBML file in same folder as python script
                f.write(sbmlstr)
                f.close()
                del sbmlstr
                sbmlfilelist.append(sbmlfilename)
    
    print()
    print('Number of Files Printed =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)

                            



def autogenerate_sbml_from_folder(folderpath):
    '''To automatically generate the corresponding SBML files from the given
    folder containing all the configuration files.
    
    :param folderpath: the path to the folder containing all the configuration files.
    :type folderpath: str
    '''
        
    files = [f for f in glob.glob(os.path.join(folderpath,"**/*.ini"), recursive=True)]
    
    for f in files:
        print('\n', f)

    sbmlfilelist=[]
    number_scenario = 0
       
    for f in files:    #Add ini files to database
        system_type = mh.config_to_database(f)
        system_types_settings_names = sh.config_to_database(f)
        system_type, settings_name = system_types_settings_names[0]
        search_result_model = mh.quick_search(system_type)
        search_result_settings = sh.quick_search(system_type=system_type, settings_name=settings_name)
        core_model    = search_result_model
        settings = search_result_settings
        addparam =  settings['parameters']
        number_init = len(settings['init'])
        number_parameters = len(settings['parameters'])
        unit_model = unitlookup(settings)
       
        for j in range(number_init): #Cycle through number of init values
            for k in range(number_parameters): #Cycle through number of parameters
                sbmlstr = SBMLcreation(core_model, settings, unit_model, addparam, j, k)
                number_scenario = number_scenario + 1
                placeholder = os.path.splitext(os.path.basename(f))[0]
                sbmlfilename = os.path.join(folderpath, 'DatabasetoSBML_' + placeholder +'.xml')
                f = open(sbmlfilename, 'w') #creates SBML file in same folder as python script
                f.write(sbmlstr)
                f.close()
                del sbmlstr
                sbmlfilelist.append(sbmlfilename)
        
        #Function that creates SBML file and returns the number of files outputed and the file list.
        
    
    print()
    print('Number of Files Printed =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)


def database_to_sbml(system_type, settings_name):
    '''Select model stored in database and generate the SBML file. 
    '''
    sbmlfilelist    = []
    number_scenario = 0

    
    search_result_settings = sh.search_database(system_type, settings_name)
    search_result_model    = mh.quick_search(system_type)
    core_model             = search_result_model
    settings               = search_result_settings[0]
    addparam               = settings['parameters']
    number_init            = len(settings['init'])
    number_parameters      = len(settings['parameters'])
    
    unit_model = unitlookup(settings)
    
    for j in range(number_init): #Cycle through number of init
        for k in range(number_parameters): #Cycle through number of parameters
            sbmlstr = SBMLcreation(core_model, settings, unit_model, addparam, j, k)           
            number_scenario = number_scenario + 1
            sbmlfilename = 'DatabasetoSBML_' + str(number_scenario)+'.xml'
            f = open(sbmlfilename, 'w')
            f.write(sbmlstr)
            f.close()
            del sbmlstr
            sbmlfilelist.append(sbmlfilename)
    #Function that creates SBML file and returns the number of files outputed and the file list.
        
    print()
    print('Number of Scenarios in Model =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)



import numpy             as np
from numba               import jit
from pandas              import DataFrame, Series
from scipy.optimize      import basinhopping           as sp_bh
from scipy.optimize      import differential_evolution as sp_de
from scipy.optimize      import dual_annealing         as sp_da
from scipy.stats         import norm

###############################################################################
#Sampler Algorithm
###############################################################################
def differential_evolution(guess, priors={}, skip=[], bounds={}, 
                           likelihood_function=None, 
                           likelihood_args={}, 
                           **scipy_args):
    
    '''
    Minimizes the negative of the posterior function with custom likelihood function.
    Bounds must be provided for ALL parameters
    '''

    accepted   = []
    full_trace = []
    
    temp               = convert_input(guess, priors=priors, skip=skip, bounds=bounds) 
    var_param_index    = temp['var_param_index']
    bounds_array       = temp['bounds_array']
    prior_distribution = temp['prior_distribution'] 
    prior_index        = temp['prior_index']        
    full_param_array   = temp['param_array']
    param_names        = list(guess)
    
    wrapped_posterior   = wrap_log_posterior(var_param_index, full_param_array, prior_index, prior_distribution, likelihood_function, likelihood_args, full_trace)
    wrapped_callback    = wrap_record_trace_de(var_param_index, full_param_array, accepted)
    
    res = sp_de(wrapped_posterior, bounds=bounds_array, callback=wrapped_callback, **scipy_args)
    
    accepted, full_trace = return_df(accepted, full_trace, columns=param_names)
    
    return {'a': accepted, 's': res, 'f': full_trace}
        
def basinhopping(guess, priors={}, skip=[], bounds={}, step_size=0.1, 
                 step_size_is_ratio=True,
                 likelihood_function=None, 
                 likelihood_args={}, 
                 **scipy_args):
    
    '''
    Minimizes the negative of the posterior function with custom likelihood function.
    '''
    
    accepted   = []
    rejected   = []
    full_trace = []

    temp               = convert_input(guess, priors=priors, skip=skip, bounds=bounds, step_size=step_size) 
    var_param_index    = temp['var_param_index']
    var_param_array    = temp['var_param_array']
    bounds_array       = temp['bounds_array']
    prior_distribution = temp['prior_distribution'] 
    prior_index        = temp['prior_index']        
    full_param_array   = temp['param_array']
    step_size_array    = temp['step_size_array']
    param_names        = list(guess)

    wrapped_posterior   = wrap_log_posterior(var_param_index, full_param_array, prior_index, prior_distribution, likelihood_function, likelihood_args, full_trace)
    wrapped_accept_test = wrap_bounds_test(bounds_array, *scipy_args.get('accept_test', []))
    wrapped_callback    = wrap_record_trace_bh(var_param_index, full_param_array, accepted, rejected)
    wrapped_transition  = wrap_transition(var_param_array, step_size=step_size_array, step_size_is_ratio=step_size_is_ratio)

    res = sp_bh(wrapped_posterior, x0=var_param_array, take_step=wrapped_transition, callback=wrapped_callback, accept_test=wrapped_accept_test, minimizer_kwargs={'method': 'trust-constr'}, **scipy_args)
    
    accepted, rejected, full_trace = return_df(accepted, rejected, full_trace, columns=param_names)
    
    return {'a': accepted, 'r': rejected, 's': res, 'f': full_trace}

def dual_annealing(guess,     priors={},  skip=[], bounds={},
                   likelihood_function=None, 
                   likelihood_args={}, 
                   **scipy_args):
    '''
    Minimizes the negative of the posterior function with custom likelihood function.
    '''
    accepted   = []
    contexts   = []
    full_trace = []

    temp               = convert_input(guess, priors=priors, skip=skip, bounds=bounds, bound_max=10) 
    var_param_index    = temp['var_param_index']
    var_param_array    = temp['var_param_array']
    bounds_array       = temp['bounds_array']
    prior_distribution = temp['prior_distribution'] 
    prior_index        = temp['prior_index']        
    full_param_array   = temp['param_array']
    param_names        = list(guess)
    
    wrapped_posterior   = wrap_log_posterior(var_param_index, full_param_array, prior_index, prior_distribution, likelihood_function, likelihood_args, full_trace)
    wrapped_callback    = wrap_record_trace_da(var_param_index, full_param_array, accepted, contexts)

    res = sp_da(wrapped_posterior, bounds=bounds_array, x0=var_param_array, callback=wrapped_callback, **scipy_args)
    
    accepted, full_trace = return_df(accepted, full_trace, columns=param_names)

    return {'a': accepted, 'c': contexts, 'f': full_trace, 's': res}

def return_df(*arrays, columns):
    result = []
    for array in arrays:
        if len(array):
            result.append( DataFrame(np.array(array), columns=columns) )
        else:
            result.append( DataFrame(columns=columns) )
    return result
            
###############################################################################
#Preprocessing
###############################################################################
def convert_input(guess, skip, bounds={}, priors={}, step_size=0.1, bound_max='inf'):
    
    check_input(guess, priors, step_size, skip, bounds)
    print('guess', guess)
    result             = {}
    param_array        = params_to_array(guess)
    print('param_array', param_array)
    param_names        = list(guess) if type(guess) == dict else list(guess.columns)
    name_to_num        = {name: num for num, name in enumerate(param_names)}
    print('n2n', name_to_num)
    var_param_index    = np.array([name_to_num[name] for name in guess if name not in skip])
    print('var_param_index', var_param_index)
    var_param_array    = param_array[var_param_index]
    max_val            = {name: np.inf for name in param_names} if bound_max == 'inf' else {pair[0]:max(pair[1], 1e-7) for pair in zip([name for name in guess if name not in skip], bound_max*var_param_array)}
    bounds_array       = np.array([bounds.get(name, [0, max_val[name]]) for name in guess if name not in skip])
    prior_distribution = get_prior_distribution(priors) if priors else None
    prior_index        = get_prior_index(guess, priors) if priors else []

    if type(step_size) == dict:
        step_size1      = {}
        for key in guess:
            if key in skip:
                continue
            elif key in step_size:
                step_size1[key] = step_size[key] 
            elif 'n' != key[0]:
                step_size1[key] = guess[key]/15 
            else:
                step_size1[key] = guess[key]/10 
        step_size_array = np.array(list(step_size1.values()))
    else:
        step_size_array = step_size 

    result['param_array']        = param_array
    result['var_param_index']    = var_param_index
    result['var_param_array']    = var_param_array
    result['bounds_array']       = bounds_array
    result['prior_distribution'] = prior_distribution
    result['prior_index']        = prior_index  
    result['step_size_array']    = step_size_array
    
    return result

def params_to_array(params):
    if type(params) == np.ndarray:
        return params
    elif type(params) == dict:
        try:
            temp = params[next(iter(params))]
            next(iter(temp))
            raise Exception('Only one value of params allowed!')
        except:
            result = np.array(list(params.values()))
        
            return np.array(result)
    elif type(params) == DataFrame:
        if params.shape[0] > 1:
            raise Exception('Only one value of params allowed!')
        return params.to_numpy()
    elif type(params)  == Series:
        return params.values
    else:
        raise Exception('Params is neither np.ndarry, dict nor DataFrame.')
        
def get_prior_index(params, priors):
    return np.array([pair[0] for pair in enumerate(params) if priors.get(pair[1], False)])

def get_prior_distribution(priors):  
    values      = np.array(list(priors.values()))   
    return norm(values[:,0], values[:,1])

def check_input(guess, priors, step_size={}, skip=[], bounds={}):
    if type(guess) != dict and type(guess) != Series: 
        raise Exception('Guess is neither dictionary nor Series.')
    
    guess_keys = guess.keys() if type(guess) == dict else list(guess.index)
    
    if type(step_size) == dict:
        for key in step_size:
            if key not in guess_keys:
                raise Exception('Key ' + key + ' in step_size not in present in guess_keys.')
    else:
        try:
            len(step_size)
        except:
            pass
        else:
            raise Exception('step_size must be dict or number.')
        
    for key in priors:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in priors not in present in guess_keys.')
    
    for key in bounds:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in bounds not in present in guess_keys.')
        if len(bounds[key]) != 2:
            raise Exception('Bounds[' + key + '] must be of length 2: [lower_bound, upper_bound]')
        if bounds[key][0] > bounds[key][1]:
            raise Exception('Bounds[' + key + ']: lower_bound must be smaller than upper_bound.')
        if guess[key] < bounds[key][0] or guess[key] > bounds[key][1]:
            raise Exception('Value of guess outside bounds for key: ' + key)
            
###############################################################################
#Posterior Calculation and Acceptance
###############################################################################
def wrap_log_posterior(var_param_index, full_params_array, prior_index, prior_distribution, likelihood_function, likelihood_args, trace=[]):
    def helper(var_param_array):
        log_posterior, log_prior, log_likelihood = get_negative_log_posterior(var_param_array, var_param_index, full_params_array, prior_index, prior_distribution, likelihood_function, likelihood_args, trace)
        return log_posterior
    return helper

def get_negative_log_posterior(var_param_array, var_param_index, full_params_array, prior_index, prior_distribution, likelihood_function, likelihood_args, trace=[]):
    params_array   = reconstruct_params(var_param_array, var_param_index, full_params_array)
    log_prior      = get_log_prior(params_array[prior_index], prior_distribution)
    log_likelihood = get_log_likelihood(params_array, likelihood_function=likelihood_function, likelihood_args=likelihood_args)
    log_posterior  = -(log_prior + log_likelihood) 
    
    trace.append(params_array)
    
    return log_posterior, log_prior, log_likelihood

@jit(nopython=True)
def reconstruct_params(variable_params, variable_param_index, params_array):
    params_array1 = params_array.copy()
    
    params_array1[variable_param_index] = variable_params
    
    return params_array1

###############################################################################
#Likelihood and Prior Calculation
###############################################################################
def get_log_likelihood(params_array, likelihood_function, likelihood_args={}):
    return likelihood_function(params_array=params_array, **likelihood_args)

def get_log_prior(params={}, priors={}):
    '''
    Pass in a dicts of params and priors OR np.arrays of params and priors
    '''
    if not priors:
        return 0
    
    elif type(priors) == dict: 
        #For testing purposes
        prior_distribution = get_prior_distribution(priors)
        params_array       = np.array([params[key] for key in priors.keys()])
        return get_log_sum(prior_distribution.pdf(params_array))
        
    else: 
        #Assumes params is already an array and has been sliced appropriately
        #Assumes priors is already a ditribution
        return get_log_sum(priors.pdf(params))

@jit(nopython=True)
def get_log_sum(a):
    return np.log(a).sum()

###############################################################################
#Transition
###############################################################################
def wrap_transition(guess_array, step_size=0.1, step_size_is_ratio=True):
    '''
    Generates a scipy.norm object for generating random variables.
    Wraps the transition step into a function that accepts the current array and returns the transitioned one.
    Use the get_step argument to return the norm object instead.
    If size_type_is_ratio is True, the sd equal to guess*step.
    If False, the sd equal to step.
    '''
    if step_size_is_ratio:
        step = norm([0]*len(guess_array), guess_array*step_size)
    else:
        step = norm([0]*len(guess_array), step_size)
        
    def helper(var_param_array, get_step=False):
        if not get_step:
            delta     = step.rvs()
            new_array = var_param_array + delta
            return new_array
        else:
            return step
    return helper

###############################################################################
#Tracing
###############################################################################
def wrap_record_trace_bh(var_param_index, params_array, accepted, rejected):
    def helper(x, f, accept):
        if accept:
            record_trace(x, var_param_index, params_array, accepted)
        else:
            record_trace(x, var_param_index, params_array, rejected)
        return
    return helper

def wrap_record_trace_da(var_param_index, params_array, accepted, contexts):
    def helper(x, f, context):
        
        record_trace(x, var_param_index, params_array, accepted)
        contexts.append(context)
        return
    return helper

def wrap_record_trace_de(var_param_index, params_array, trace):
    def helper(xk, convergence):
        record_trace(xk, var_param_index, params_array, trace)
        return
    return helper
               
def record_trace(var_param_array, var_param_index, params_array, trace):
    full_params_array = reconstruct_params(var_param_array, var_param_index, params_array)
    
    return trace.append(full_params_array)

###############################################################################
#Enforcing Bounds
###############################################################################    
def wrap_bounds_test(bounds_array, *funcs):
    '''
    Wraps bounds_test as well as any custom functions.
    '''
    def helper(f_new, x_new, f_old, x_old):
        var_param_array = x_new

        if not bounds_test(var_param_array, bounds_array):
            return False
        else:
            for func in funcs:
                if not func(f_new, x_new, f_old, x_old):
                    return False
            return True
    return helper

def bounds_test(var_param_array, bounds_array):
    if not np.all(var_param_array >= bounds_array[:,0]):
        return False
    elif not np.all(var_param_array <= bounds_array[:,1]):
        return False
    else:
        return True
    

def dummy(params_array, **kwargs):
    result = np.sum((50-params_array)**2)
    return result

if __name__ == '__main__':
    guess = {'p'+str(i):100 for i in range(21)}
    skip  = []
    
    priors = {}
    likelihood_function = dummy
    likelihood_args = {}
    scipy_args = {'niter':100}
    
    accepted = []
    rejected = []

    temp               = convert_input(guess, priors=priors, skip=skip, bounds={}) 
    var_param_index    = temp['var_param_index']
    var_param_array    = temp['var_param_array']
    bounds_array       = temp['bounds_array']
    prior_distribution = temp['prior_distribution'] 
    prior_index        = temp['prior_index']        
    full_param_array   = temp['param_array']
    param_names        = list(guess)
    
    wrapped_posterior   = wrap_log_posterior(var_param_index, full_param_array, prior_index, prior_distribution, likelihood_function, likelihood_args)
    
    wrapped_accept_test = wrap_bounds_test(bounds_array, *scipy_args.get('accept_test', []))
    
    wrapped_callback    = wrap_record_trace_de(var_param_index, full_param_array, accepted, rejected)
    
    wrapped_transition  = wrap_transition(var_param_array, step_size=10, step_size_is_ratio=False)
    




import numpy             as np
from numba               import jit
from pandas              import DataFrame
from scipy.stats         import norm
from scipy.integrate     import odeint

###############################################################################
#Sampler Algorithm
###############################################################################
def sampler(guess,     priors,    step_size, 
            trials=10000, skip=[], bounds={}, blocks=[], SA=True, 
            likelihood_function=None, 
            likelihood_args={}):
    
    accepted = []
    rejected = []
    
    temp               = convert_input(guess, priors, step_size, skip=skip, bounds=bounds, blocks=blocks) 
    blocks_dict        = temp['blocks_dict']
    bounds_dict        = temp['bounds_dict']
    params_array       = temp['params_array']
    param_names        = temp['param_names']
    prior_distribution = temp['prior_distribution']
    prior_index        = temp['prior_index']
    step_size_rvs      = temp['step_size_rvs']

    log_prior_curr      = None
    log_likelihood_curr = None
    p_curr              = None
    
    theta_curr  = params_array.copy()
    block_index = 0
    for i in range(trials):
        #Transition
        block            = blocks_dict[block_index]
        theta_new        = theta_curr.copy()
        theta_new[block] = transition(theta_curr[block], step_size_rvs[block])

        block_index += 1

        if block_index == len(blocks_dict):
            block_index  = 0
        
        #Check bounds
        skip = any(param < 0 for param in theta_new)
        
        for index in bounds_dict:
            lb, ub = bounds_dict[index]
            if theta_new[index] > ub or theta_new[index] < lb:
                skip = True
                break

        if skip:
            skip = False
            step = theta_new
            rejected.append(step)
            continue
        
        #Calculate posterior
        if log_prior_curr == None:
            p_curr, log_prior_curr, log_likelihood_curr = get_log_posterior(theta_curr, prior_index, prior_distribution, likelihood_function, likelihood_args)
        
        p_new, log_prior_new, log_likelihood_new = get_log_posterior(theta_new, prior_index, prior_distribution, likelihood_function, likelihood_args)
                
        #Calculate acceptance
        p_accept = get_exp_substract(p_new, p_curr) 
        
        #Compare with random number
        test   = np.random.rand()
        if SA == True:
            test = test**(1-1*i/trials)
        accept = p_accept > test
        
        #Collect sample
        if accept:
            theta_curr          = theta_new
            log_likelihood_curr = log_likelihood_new
            log_prior_curr      = log_prior_new
            p_curr              = p_new
            accepted.append(theta_new)
 
        else:
            rejected.append(theta_new)
    
    accepted = DataFrame(np.array(accepted), columns=param_names) if accepted else DataFrame(columns=param_names)
    rejected = DataFrame(np.array(rejected), columns=param_names) if rejected else DataFrame(columns=param_names)
    return {'a': accepted, 'r': rejected}

@jit(nopython=True)
def get_exp_substract(a, b):
    return np.exp(a-b)

###############################################################################
#Conversion
###############################################################################    
def check_input(guess, priors, step_size={}, skip=[], bounds={}, blocks=[]):
    if type(guess) == dict:
        guess_keys = guess.keys()
            
    elif type(guess) == DataFrame:
        guess_keys = guess.columns
    else:
        raise Exception('Guess is neither dictionary nor DataFrame.')
    
    if not all([type(key) == str for key in guess_keys]):
        raise Exception('Parameter names must be strings.')
        
    for key in step_size:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in step_size not in present in guess_keys.')
    
    for key in priors:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in priors not in present in guess_keys.')
    
    for key in skip:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in skip not in present in guess_keys.')
    
    for key in bounds:
        if key not in guess_keys:
            raise Exception('Key ' + key + ' in bounds not in present in guess_keys.')
        if len(bounds[key]) != 2:
            raise Exception('Bounds[' + key + '] must be of length 2: [lower_bound, upper_bound]')
        if bounds[key][0] > bounds[key][1]:
            raise Exception('Bounds[' + key + ']: lower_bound must be smaller than upper_bound.')
        if guess[key] < bounds[key][0] or guess[key] > bounds[key][1]:
            raise Exception('Value of guess outside bounds for key: ' + key)
            
    for block in blocks:
        for key in block:
            if key not in guess_keys:
                raise Exception('Key ' + key + ' in blocks not in present in guess_keys.')


def convert_input(guess, priors, step_size, skip=[], bounds={}, blocks=[]):
    
    check_input(guess, priors, step_size, skip=skip, bounds=bounds, blocks=blocks)
    
    guess_dict         = params_to_dict(guess)
    params_array       = params_to_array(guess_dict)[0]
    prior_distribution = get_prior_distribution(priors) if priors else None
    prior_index        = get_prior_index(guess, priors) if priors else []
    param_names        = list(guess_dict.keys())
    name_to_num        = {name: num for num, name in enumerate(param_names)}
    skip_index         = [name_to_num[name] for name in skip]
    blocks_array       = [[name_to_num[name] for name in block ] for block in blocks ]
    blocks_dict        = {}
    bounds_dict        = {name_to_num[name]: bounds[name] for name in bounds}
    step_size1         = {key: step_size[key] if key in step_size else guess[key]/15 if 'n' != key[0] else guess[key]/10 for key in guess_dict}
    step_size_rvs      = np.array([norm(0, step_size1[key]).rvs for key in step_size1])
    
    n = 0
    for i in range(len(guess)):
        if i in skip_index:
            continue
        
        in_block = False
        for block in blocks_array:
            if i in block:
                blocks_dict[n] = block
                in_block       = True
                break
        if not in_block:
            blocks_dict[n] = [i]
        n += 1
    
    result = {'blocks_array'      : blocks_array,
              'blocks_dict'       : blocks_dict,
              'bounds_dict'       : bounds_dict,
              'params_array'      : params_array,
              'param_names'       : param_names,
              'prior_distribution': prior_distribution,
              'prior_index'       : prior_index,
              'skip_index'        : skip_index,
              'step_size_rvs'     : step_size_rvs,
              }
    return result

def get_default_step_size(guess):
    if type(guess) == dict:
        return {key: guess[key]/15 if 'n' != key[0] else guess[key]/10 for key in guess}
    else:
        raise Exception('Guess is not a dict.')

def params_to_dict(params):
    if type(params) == dict:
        return params
    if type(params) == DataFrame:
        params_dict = params.to_dict('list')
        #Ensure that values are not enclosed in lists
        for key in params_dict:
            if type(params_dict[key]) == list:
                params_dict[key] = params_dict[key][0] 
        return params_dict
    else:
        raise Exception('Params must be a dict or DataFrame.')
        
def params_to_array(params):
    if type(params) == np.ndarray:
        return params
    elif type(params) == dict:
        result = np.array([row for row in params.values()])
        if type(result[0]) == np.ndarray:
            return result.T
        else:
            return np.array([result])
    elif type(params) == DataFrame:
        return params.to_numpy()
    else:
        raise Exception('Params is neither np.ndarry, dict nor DataFrame.')
 
###############################################################################
#Transition
###############################################################################
def transition(theta, step_size_rvs):
    delta = np.array([rvs() for rvs in step_size_rvs])
    return theta + delta

###############################################################################
#Posterior Calculation and Acceptance
###############################################################################
def get_log_posterior(theta, prior_index, prior_distribution, likelihood_function, likelihood_args):
    log_prior      = get_log_prior(theta[prior_index], prior_distribution)
    log_likelihood = get_log_likelihood(theta, likelihood_function=likelihood_function, likelihood_args=likelihood_args)
    
    log_posterior  = log_prior + log_likelihood

    return log_posterior, log_prior, log_likelihood

###############################################################################
#Likelihood and Prior Calculation
###############################################################################
def get_log_likelihood(params_array, likelihood_function=None, likelihood_args={}):
    return likelihood_function(params_array=params_array, **likelihood_args)

def get_log_prior(params={}, priors={}):
    '''
    Pass in a dicts of params and priors OR np.arrays of params and priors
    '''
    if not priors:

        return 0
    
    elif type(priors) == dict:  
        if type( priors[next(iter(priors))] ) == list:
            prior_distribution = get_prior_distribution(priors)
            params_array       = np.array([params[key] for key in priors.keys()])
            return get_log_sum(prior_distribution.pdf(params_array))
        else:
            return get_log_sum(np.array([priors[key].pdf([params[key]]) for key in priors]))
        
    else: 
        #Assumes params is already an array and has been sliced appropriately
        #Assumes priors is already a ditribution
        return get_log_sum(priors.pdf(params))

def get_prior_index(params, priors):
    return np.array([pair[0] for pair in enumerate(params) if priors.get(pair[1], False)])

def get_prior_distribution(priors):  
    values      = np.array(list(priors.values()))   
    return norm(values[:,0], values[:,1])

@jit(nopython=True)
def get_log_sum(a):
    return np.log(a).sum()


import matplotlib.pyplot as plt 
import numpy             as np
import pandas            as pd
import seaborn           as sns
from matplotlib          import get_backend
from numba               import jit
from scipy.integrate     import odeint

###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from .mcmc       import *
    from .simulation import piecewise_integrate, palette_types, all_colors, apply_titles_and_legend, fs
    from .           import scipy_wrappers as wrappers
except:
    from mcmc import *
    from simulation import piecewise_integrate, palette_types, all_colors, apply_titles_and_legend, fs
    import scipy_wrappers as wrappers
    
###############################################################################
#SSE Calculation
###############################################################################
def get_SSE_data(data, models, params={}, t_indices={}, params_array=np.array([]), params_index={}):
    '''
    Calculates SSE for data against models and returns the negative of the SSE.
    
    For testing/single calls:
        1. Supply the parameters as a dict. Ignore all other optional arguments.
    For iterative calls:
        1. Ignore the params argument.
        2. Provide a numpy array of parameters instead.
        3. Use the function get_liklihood_args to set up the other arguments.

    Parameters
    ----------
    data : dict
        Experimental data.
    models : dict
        Dictionary of model data structures.
    params : dict, optional
        For backend use only. The default is {}.
    t_indices : dict, optional
        For backend use only. The default is {}.
    params_array : nupy.array, optional
        An array of parameter values. The default is np.array([]).
    params_index : dict, optional
        Required to work with the params_array. The default is {}.

    Returns
    -------
    float
        The negative of the SSE.

    '''
    if params:
        params_array1  = np.array(list(params.values()))
        name_to_num   = {name: num for num, name in enumerate(params)}
        params_index1  = {model_num: [name_to_num[param] for param in models[model_num]['params']] for model_num in models}
    else:
        params_array1 = params_array
        params_index1 = params_index
    
    t_indices1 = t_indices if t_indices else get_t_indices(data, models)
        
    SSE  = 0
    
    for model_num in data:
        
        params_  = params_array1[params_index1[model_num]] 
        SSE     += get_SSE_dataset(model_num, data, models, params_, t_indices1)

    return -SSE #Negated as required by likelihood function

###############################################################################
#Supporting Functions for SSE Calculation. Do not run.
###############################################################################
def get_SSE_dataset(model_num, data, models, params, t_indices):
    '''
    :meta private:
    '''
    SSE = 0

    dataset  = data[model_num]
    model    = models[model_num]
    tspan    = get_tspan(model)
    init     = model['init']
    int_args = model.get('int_args', {})
    
    for scenario in range(1, len(init)+1):
        y_model, t_model  = piecewise_integrate(model['function'], init[scenario], tspan, params, model_num, scenario, **int_args, overlap=False)
        
        for i in range(len(model['states'])):
            state      = model['states'][i]
            if state not in dataset:
                continue
            SSE_column = i
            y_data     = dataset[state][scenario]
            indices    = t_indices[model_num][state]
            y_model_   = y_model[indices,SSE_column] 
            sd_dataset = model['sd'][state]
            SSE += get_SSE(y_data, y_model_, sd_dataset)

    return SSE

@jit(nopython=True)
def get_SSE(y_data, y_model, sd_dataset=1):
    '''
    :meta private:
    '''
    error  = (y_data - y_model)**2
    return error.sum()/sd_dataset**2 

def get_t_indices(data, models):
    '''
    :meta private:
    '''
    t_indices = {}
    for model_num in data:
        dataset, model         = data[model_num], models[model_num]
        t_indices[model_num] = get_indices_dataset(dataset, model)
    return t_indices
    
def get_indices_dataset(dataset, model):
    '''
    :meta private:
    '''
    indices = {}
    for state in dataset:
        t_model            = flatten_tspan(model['tspan'])
        indices[state] = get_indices(dataset[state][0], t_model)
    return indices
        
def get_indices(t_data, t_model):
    '''
    :meta private:
    '''
    indices = []
    for t in t_data:
        index = np.where(t_model == t)

        if len(index) > 0:
            if len(index[0]) > 0:
                indices.append(int(index[0]))
            else:
                raise Exception(str(t)+' not in t_model.')
    return indices

def flatten_tspan(tspan):
    '''
    :meta private:
    '''
    return [t for segment in tspan for t in segment]

###############################################################################
#Set Up Objective Args
###############################################################################
def get_likelihood_args(data, models, params):
    '''
    Meant to be used with get_SSE. Accepts data, models and a dict of params to 
    generate the remaining arguments for get_SSE.
    
    Note: Actual values of the params do not matter. Only the order.
    
    :meta private:
    '''
    name_to_num   = {name: num for num, name in enumerate(params)}
    params_index  = {model_num: [name_to_num[param] for param in models[model_num]['params']] for model_num in models}
    t_indices     = get_t_indices(data, models)
    result        =  {'data'        : data, 
                      'models'      : models, 
                      't_indices'   : t_indices, 
                      'params_index': params_index
                      }
    return result        

###############################################################################
#Main Algorithm
###############################################################################
def simulated_annealing(data,         models,    guess,     priors,    step_size, 
                        trials=10000, bounds={}, blocks=[], SA=True, 
                        fixed_parameters=[],
                        likelihood_function=get_SSE_data,
                        **kwargs
                        ):  
    
    '''
    BMSS2's primary curve-fitting algorithm.
    
    Parameters
    ----------
    data : dict
        Experimental data.
    models : dict
        Dictionary of model data structures.
    guess : dict
        The initial guess. The default is {}.
    priors : dict, optional
        Priors if any. The default is {}.
    step_size : dict, optional
        A dictionary mapping parameter names to median step-size during the 
        stochastic transition step.
    trials : int, optional
        The number of iterations. The default is 10000.
    bounds : dict, optional
        A dictionary mapping parameter names to tuples (lower, upper). The 
        default is {}.
    blocks : list of lists, optional
        A list of lists grouping parameter names into blocks for blocked Gibbs 
        sampling
    SA : bool, optional
        True if you want to apply the temperature criterion. False if you want 
        to use a constant cutoff criterion. The default is True.
    fixed_parameters : list, optional
        A list of parameters that will be fixed. The default is [].
    likelihood_function : function, optional
        The likelihood function to be calculated. Do not change unless you know 
        what you are doing.
    
    Returns
    ---------- 
    result : dict
        A dictionary with the following mapping:
            "a" : The accepted samples
            "r" : The rejected samples
    '''
    
    likelihood_args = get_likelihood_args(data, models, guess)
    
    return sampler(guess,         priors,        step_size, 
                   trials=trials, skip=fixed_parameters, bounds=bounds, blocks=blocks, SA=SA, 
                   likelihood_function=get_SSE_data,
                   likelihood_args=likelihood_args,
                   **kwargs)

def scipy_basinhopping(data,    models,    guess, priors, step_size=0.1, bounds={},  
                       step_size_is_ratio=True,
                       fixed_parameters=[],
                       likelihood_function=get_SSE_data, 
                       scipy_args={},
                       **kwargs):
    '''
    Wrapper for scipy's basin-hopping algorithm.
    
    Parameters
    ----------
    data : dict
        Experimental data.
    models : dict
        Dictionary of model data structures.
    guess : dict
        The initial guess. The default is {}.
    priors : dict, optional
        Priors if any. The default is {}.
    step_size : float or dict, optional
        Step size.
    step_size_is_ratio : bool, optional
        If True, step size is a proportion of the initial guess. If False, 
        absolute values are used.
    bounds : dict, optional
        A dictionary mapping parameter names to tuples (lower, upper). The 
        default is {}.
    fixed_parameters : list, optional
        A list of parameters that will be fixed. The default is [].
    likelihood_function : function, optional
        The likelihood function to be calculated. Do not change unless you know 
        what you are doing.
    
    Returns
    ---------- 
    result : dict
        A dictionary with the following mapping:
            "a" : The accepted samples
            "r" : The rejected samples
    '''
    likelihood_args = get_likelihood_args(data, models, guess)
    
    result = wrappers.basinhopping(guess,         priors,  step_size=step_size, 
                                   skip=fixed_parameters, bounds=bounds,  
                                   likelihood_function=likelihood_function, 
                                   likelihood_args=likelihood_args,
                                   **scipy_args,
                                   **kwargs)
    
    return result

def scipy_dual_annealing(data, models, guess, priors, bounds={},  
                         fixed_parameters=[], 
                         likelihood_function=get_SSE_data, 
                         scipy_args={},
                         **kwargs):
    '''
    Wrapper for scipy's dual annealing algorithm.
    
    Parameters
    ----------
    data : dict
        Experimental data.
    models : dict
        Dictionary of model data structures.
    guess : dict
        The initial guess. The default is {}.
    priors : dict, optional
        Priors if any. The default is {}.
    bounds : dict, optional
        A dictionary mapping parameter names to tuples (lower, upper). The 
        default is {}.
    fixed_parameters : list, optional
        A list of parameters that will be fixed. The default is [].
    likelihood_function : function, optional
        The likelihood function to be calculated. Do not change unless you know 
        what you are doing.
    
    Returns
    ---------- 
    result : dict
        A dictionary with the following mapping:
            "a" : The accepted samples
            "r" : The rejected samples
    '''
    likelihood_args = get_likelihood_args(data, models, guess)
    
    result = wrappers.dual_annealing(guess,     priors, 
                                     skip=fixed_parameters, bounds=bounds,  
                                     likelihood_function=likelihood_function, 
                                     likelihood_args=likelihood_args,
                                     **scipy_args,
                                     **kwargs)
    
    return result

def scipy_differential_evolution(data, models, guess, priors, 
                                 fixed_parameters, bounds, 
                                 likelihood_function=get_SSE_data,
                                 scipy_args={}):
    '''
    Wrapper for scipy's differential evolution algorithm.
    
    Parameters
    ----------
    data : dict
        Experimental data.
    models : dict
        Dictionary of model data structures.
    guess : dict, optional
        The initial guess. The default is {}.
    priors : dict, optional
        Priors if any. The default is {}.
    bounds : dict, optional
        A dictionary mapping parameter names to tuples (lower, upper). The 
        default is {}.
    fixed_parameters : list, optional
        A list of parameters that will be fixed. The default is [].
    likelihood_function : function, optional
        The likelihood function to be calculated. Do not change unless you know 
        what you are doing.
    
    Returns
    ---------- 
    result : dict
        A dictionary with the following mapping:
            "a" : The accepted samples
    '''
    if guess:
        guess1          = guess if type(guess) == dict else guess.to_dict() 
    else:
        guess1          = {key: np.mean(bounds[key]) for key in bounds}
        
    likelihood_args = get_likelihood_args(data, models, guess1)
    
    result = wrappers.differential_evolution(guess=guess1, priors=priors, skip=fixed_parameters, bounds=bounds, 
                                             likelihood_function=likelihood_function, 
                                             likelihood_args=likelihood_args, 
                                             **scipy_args)
    
    return result

###############################################################################
#Visualization
###############################################################################
#Plot everything
def plot(posterior={},   guess={},       models={},   data={}, 
         data_sd={},     plot_index={},  titles={},   labels={},
         legend_args={}, figs=[],        AX={},       palette='color',
         line_args={'linewidth': 1.5},   guess_palette='white'):
    '''
    Wrapper for plotting function for data, guess and posterior in one function call.
    '''
    first    = True
    
    if len(posterior) > 0:
        figs1, AX1 = plot_model(models, posterior, plot_index=plot_index, titles=titles, labels=labels, legend_args=legend_args, palette=palette, figs=figs, AX=AX, line_args=line_args)
        first = False

    if len(data) > 0:
        if first:
            figs1, AX1 = plot_data(data, data_sd=data_sd, plot_index=plot_index, titles=titles, labels=labels, legend_args=legend_args, palette=palette, figs=figs, AX=AX) 
            first      = False
        else:
            figs1, AX1 = plot_data(data, data_sd=data_sd, plot_index=plot_index, titles={},     labels={},     legend_args={},          palette=palette, figs=figs1, AX=AX1) 
        
    if len(guess) > 0:
        if first:
            figs1, AX1 = plot_model(models, guess, plot_index=plot_index, titles=titles, labels=labels, legend_args=legend_args, palette=guess_palette, figs=figs,  AX=AX, line_args={'linewidth':3})
            first      = False
        else:
            figs1, AX1 = plot_model(models, guess, plot_index=plot_index, titles={},     labels={},     legend_args={},          palette=guess_palette, figs=figs1, AX=AX1, line_args={'linewidth':3})  
    
    return figs1, AX1    

def plot_model(models,         params,          plot_index={}, titles=[], labels={}, 
               legend_args={}, palette='color', figs=[],       AX={},     line_args={'linewidth': 1.5}):
    '''
    Plots the models.

    Parameters
    ----------
    models : dict
        A dictionary of model data structures.
    params : pandas.DataFrame
        A DataFrame of parameter values for integration.
    plot_index : dict
        A dictionary of states and extra variables to plot.    
    titles : dict, optional
        A dictionary of axes titles. The default is {}.
    labels : TYPE, optional
        A dictionary of line labels. The default is {}.
    figs : list of matplotlib.Figure, optional
        A list of Figure objects that will be maximized after plotting. The 
        default is ().
    AX : dict of matplotlib.Axes, optional
        A dictionary of Axes objects for plotting. If {}, the Axes will be 
        automatically generated. The default is {}.
    palette : dict, optional
        A dictionary that controls the color of the plots. The default is {}.
    line_args : dict, optional
        Keyword arguments for the plot method of the Axes class. The default is 
        {}.
    legend_args : dict, optional
        Keyword arguments for controlling the appearance of the legend. The 
        default is The default is {'linewidth': 1.5}..
        
    Returns
    -------
    figs : list
        A list of Figure objects.
    AX : dict
        A dict of Axes objects.
    '''
    figs1, AX1     = (figs, AX) if AX else make_AX(plot_index=plot_index, data={})
    params_array   = params_to_array(params)
    name_to_num    = {name: num for num, name in enumerate(params)}
    params_index   = {model_num: [name_to_num[param] for param in models[model_num]['params']] for model_num in models}
    parameter_sets = len(params_array)
    colors         = make_colors(palette=palette, models=models, parameter_sets=parameter_sets)

    for model_num in AX1:
        
        model  = models[model_num]
        tspan  = get_tspan(model)
        var    = {pair[1]: pair[0] for pair in enumerate(model['states'])}
         
        for index in range(len(params_array)):
            row         = params_array[index]
            params_     = row[params_index[model_num]]
            init        = model['init']
            int_args    = model.get('int_args', {})
            
            for scenario in range(1, len(init)+1):
                label = get(labels, model_num, scenario, no_result='')
                
                if type(label) == str:
                    label = None if index != len(params_array) - 1 else label
                else:
                    label = label[index]
                
                color = colors[model_num][scenario]
                if type(color) == dict:
                    color = colors[model_num][scenario][index]

                y, t  = piecewise_integrate(model['function'], init[scenario], tspan, params_, model_num, scenario, **int_args)
                
                for state in plot_index[model_num]:
                    ax = AX1[model_num][state]
                    ax.ticklabel_format(style='sci', scilimits=(-2,3))  
                    if type(state) in [int, str, tuple]:
                        ax.plot(t, y[:,var[state]], '-', label=label, color=color, **line_args)
                        
                    else:
                        #For extra functions
                        y_, t_, marker, kwargs = extra_state(state, y, t, params_)
                        
                        if not kwargs.get('linewidth', False):
                            kwargs['linewidth'] = line_args.get('linewidth', 1.5)
                        
                        ax.plot(t_, y_, marker, label=label, color=color, **line_args)
                     
    
    [fs(fig) for fig in figs1]      
    
    apply_titles_and_legend(AX1, titles, legend_args)
    return figs1, AX1

def plot_data(data,           data_sd={},      plot_index={}, titles={}, labels={}, 
              legend_args={}, palette='color', figs=[],       AX={},     marker='+'):  
    '''
    Plots the experimental data.

    Parameters
    ----------
    data : dict
        Experimental data.
    data_sd : dict, optional
        Standard deviation of experimental data. The default is {}.
    plot_index : dict
        A dictionary of states and extra variables to plot.    
    titles : dict, optional
        A dictionary of axes titles. The default is {}.
    labels : TYPE, optional
        A dictionary of line labels. The default is {}.
    figs : list of matplotlib.Figure, optional
        A list of Figure objects that will be maximized after plotting. The 
        default is ().
    AX : dict of matplotlib.Axes, optional
        A dictionary of Axes objects for plotting. If {}, the Axes will be 
        automatically generated. The default is {}.
    palette : dict, optional
        A dictionary that controls the color of the plots. The default is {}.
    line_args : dict, optional
        Keyword arguments for the plot method of the Axes class. The default is 
        {}.
    legend_args : dict, optional
        Keyword arguments for controlling the appearance of the legend. The 
        default is The default is {'linewidth': 1.5}.
    marker : TYPE, optional
        DESCRIPTION. The default is '+'.
        
    Returns
    -------
    figs : list
        A list of Figure objects.
    AX : dict
        A dict of Axes objects.
    '''
    figs1, AX1 = (figs, AX) if AX else make_AX(plot_index=plot_index, data=data)
    colors     = make_colors(palette=palette, data=data)
    
    for model_num in AX1:
        for state in plot_index[model_num]:
            ax = get(AX1, model_num, state, no_result=None)
            ax.ticklabel_format(style='sci', scilimits=(-2,3))  
            if ax:         
                if type(state) in [int, str, tuple]:
                    state_data = get(data, model_num, state, no_result=None)
                    if not state_data:
                        continue
                    time          = state_data[0]
                    
                    for scenario in state_data:
                        if scenario > 0:
                            y     = state_data[scenario]
                            label = get(labels, model_num, scenario, no_result='')
                            
                            color = colors[model_num][scenario]
                            if type(color) == dict:
                                color = color[next(iter(color))]
                                
                            if data_sd:
                                sd = data_sd[model_num][state][scenario]
                                ax.errorbar(time, y, yerr=sd, marker=marker, linestyle='', label=label, markersize=10, markeredgewidth=2, color=color)
                            else:
                                ax.plot(time, y, marker, label=label, markersize=10, markeredgewidth=2, color=color)
            else:
                pass
    

    [fs(fig) for fig in figs1]  
    apply_titles_and_legend(AX1, titles, legend_args)
        
    return figs1, AX1

def make_AX(plot_index={}, data={}):
    '''Generates a dictionary of axes objects using either plot_index OR data.
    '''
    dataset_keys = plot_index if plot_index else {model_num: list(data[model_num].keys()) for model_num in data}
    max_len      = 0

    for model_num in dataset_keys:
        max_len = max(max_len, len(plot_index[model_num]))
    
    figs = [plt.figure() for i in range(max_len)]
    AX_  = {}
    AX   = {}
    for i in range(max_len):
        fig = figs[i]
        cols = 0
        for model_num in dataset_keys:
            try:
                key           = dataset_keys[model_num][i]
                cols += 1
            except:
                pass
        AX_[i] = [fig.add_subplot(1, cols, i+1) for i in range(cols)]
        
        c = 0
        for model_num in dataset_keys:
            try:
                key           = dataset_keys[model_num][i]
                
                if model_num not in AX:
                    AX[model_num] = {}
                    
                AX[model_num][key] = AX_[i][c]
 
                c += 1
            except:
                pass
    return figs, AX

###############################################################################
#Supporting Functions for Visualization. Do not run.
###############################################################################    
def extra_state(func, y, t, params):
    '''
    :meta private:
    '''
    #return y_, t_, marker, kwargs
    result = func(y, t, params)
    if len(result) == 2:
        return result[0], result[1], '-', {}
    elif len(result) == 3 and type(result[2]) == str:
        return result[0], result[1], result[2], {}
    elif len(result) == 3:
        return result[0], result[1], '-', result[2]
    elif len(result) == 4:
        return result
    else:
        raise Exception('Wrong output for extra state: ' + ','.join([type(x) for x in result]))

def view_palette(palette):
    '''
    :meta private:
    '''
    return sns.palplot(palette)

###############################################################################
#Color Palette Generation
###############################################################################
def make_palette(n=0, model={}, dataset={}, palette_type='color', **kwargs):
    '''
    :meta private:
    '''
    global palette_types
    global all_colors
    func = palette_types.get(palette_type)
    
    if model:
        n_scenarios = len(model['init'])
        if func:
            palette = func(n_scenarios, **kwargs)
            palette = palette
        else:
            palette = [all_colors[palette_type]]*n_scenarios
    elif dataset:
        state_data = dataset[next(iter(dataset))]
        n_scenarios = len([key for key in state_data if key > 0])

        if func:
            palette = func(n_scenarios, **kwargs)
            palette = palette
        else:
            palette = [all_colors[palette_type]]*n_scenarios
    else:
        if func:
            palette = func(n, **kwargs)
            palette = palette
        else:
            palette = [all_colors[palette_type]]*n_scenarios
            
    return palette       

def make_colors(palette, models={}, data={}, parameter_sets=0):
    '''
    :meta private:
    '''
    if type(palette) == dict:
        if type(next(iter(palette))) == int:
            #Assumes dictionary mapping
            return palette
        elif 'primary' in palette and models:
            result_ = make_colors(palette['primary'], models=models, parameter_sets=0)
            return add_secondary_palette(result_, palette['secondary'], parameter_sets)
        else:
            kwargs = palette
    elif type(palette) == str:
        kwargs = {'palette_type': palette}
    
    result = {}
    
    if models:
        for model_num in models:
            result[model_num] = {}
                    
            palette_            = make_palette(model=models[model_num], **kwargs)

            for scenario in range(1, len(models[model_num]['init'])+1):
                result[model_num][scenario] = palette_[scenario-1]
               
    if data:
        for model_num in data:
            result[model_num] = {}
            palette_            = make_palette(dataset=data[model_num], **kwargs)
            state_data       = data[model_num][next(iter(data[model_num]))]
            for scenario in state_data:
                if scenario > 0:
                    result[model_num][scenario] = palette_[scenario-1]

    return result

def add_secondary_palette(palette, secondary, parameter_sets):
    '''
    :meta private:
    '''
    kwargs = secondary.copy()
    for model_num in palette:
        for scenario in palette[model_num]:
            kwargs['color'] = palette[model_num][scenario]
        
            sec_palette                    = make_palette(n=parameter_sets, **kwargs)
            palette[model_num][scenario] = sec_palette
    return palette
      
def get(nested_dict, first_key, second_key, no_result=None):
    '''
    :meta private:
    '''
    try:
        return nested_dict[first_key][second_key]
    except:
        return no_result
    
def get_tspan(model):
    '''
    :meta private:
    '''
    result = model.get('tspan')
    if result:
        if type(result[0]) == np.ndarray:
            return result
        else:
            return np.array( [result] )
    else:
        raise Exception('Missing tspan')
        
###############################################################################
#Data Export
###############################################################################
def save_plots(file_name, figures):
    names = []
    if type(figures) == dict:
        for figure_type in figures:
            figure_name = file_name + '_' + figure_type
            if type(figures[figure_type]) == list:
                
                for figure in range(len(figures[figure_type])):
                    name = figure_name+'_'+str(figure + 1)
                    figures[figure_type][figure].savefig(name)
                    names.append(name)
            else:
                figures[figure_type].savefig(figure_name)
                names.append(figure_name)
    
    else:
        [figures[i].savefig(file_name + '_' + str(i+1)) for i in range(len(figures))]
        names = [file_name + '_' + str(i+1) for i in range(len(figures))]
    return names

###############################################################################
#Trace Handling
###############################################################################
def get_params_for_model(models, trace, model_num, row_index):
    '''
    Extracts the parameters for a particular model using the .loc method for 
    pandas.DataFrame.

    Parameters
    ----------
    models : dict
        Dictionary of model data structures.
    trace : pandas.DataFrame
        The accepted samples.
    model_num : int
        The number of the model you are interested in.
    row_index : int
        The row in the trace you want.

    Returns
    -------
    pandas.Series or pandas.DataFrame
        The parameters for the model.

    '''
    param_names = models[model_num]['params']
    return trace.loc[row_index, param_names]

# ###############################################################################
# #Setup and Working with TimeSeries
# ###############################################################################
# def extract_from_timeseries(timeseries_dict, states, init={}, keys=[], start=0, stop=None, interval=1, skip=[], subtract_blank=True, subtract_initial=False):
#     '''
#     Accepts a dict of BMSS.timeseries.timeseries objects and performs preprocessing.
#     Includes truncation, thinning, skipping variables and blank subtraction.
#     Returns a dict with arguments required for running curvefitting.
#     '''
#     CFS                  = import_timeseries(timeseries_dict, keys=keys, start=start, stop=stop, interval=interval, skip=skip, subtract_blank=subtract_blank, subtract_initial=subtract_initial)    
#     all_data             = {}
#     all_data['init']     = make_init(states=states, data_dict=CFS['mu'], init=init)
#     all_data['data']     = get_states(states=states, data_dict=CFS['data'])
#     all_data['mu']       = get_states(states=states, data_dict=CFS['mu'])
#     all_data['sd']       = get_states(states=states, data_dict=CFS['sd'])
#     all_data['tspan']    = make_tspan(data_dict = CFS['data'])
#     all_data['state_sd'] = CFS['state_sd']
    
#     return all_data

# def import_timeseries(timeseries_dict, keys=[], start=0, stop=None, interval=1, skip=[], subtract_blank=True, subtract_initial=False):
#     '''
#     Supporting function for extrac_from_timeseries. Do not run.
#     '''
#     temp            = ts.unpack(timeseries_dict, keys=keys, dtype='curvefit', start=start, stop=stop, interval=interval, skip=skip, subtract_blank=subtract_blank, subtract_initial=subtract_initial)
#     CFS             = {}
#     CFS['data']     = {state: temp[state]['data'] for state in temp}
#     CFS['mu']       = {state: temp[state]['mu'] for state in temp}
#     CFS['sd']       = {state: temp[state]['sd'] for state in temp}
#     CFS['state_sd'] = extract_state_sd(CFS['sd'])
    
#     return CFS

# def extract_state_sd(CFS_sd):
#     '''
#     Supporting function for extrac_from_timeseries. Do not run.
#     '''
#     sd = {state: np.max( [ np.max(CFS_sd[state][key]) for key in CFS_sd[state] if key > 0] ) for state in CFS_sd}
#     return sd
    
# def make_init(states, init={}, data_dict={}, width=1):
#     '''
    
#     Supporting function for extrac_from_timeseries. Do not run.
    
#     Creates a dictionary of initial values for each state.
#     If data_dict is provided, this function returns the first value of the array for each key that is more than 0. 
#     Values in init that are numbers will be concatenated to the appropriate length.
#     Values in init that are list-like will be used as-is but must be of the correct length.
#     If data_dict is not provided, width will be used to control the array size.
#     Width is ignored when data_dict is provided.
#     Values in init will always override those in data_dict.
#     '''
#     first_value = {state:[] for state in states}
#     for state in data_dict:
#         d = data_dict[state]
            
#         first_value[state] = np.array([d[key][0] for key in d if key > 0])
    
#     n = len(first_value[next(iter(first_value))]) if data_dict else width
    
#     for state in init:
#         try:
#             next(iter(init[state]))
#             first_value[state] = np.array(init[state])
#         except:
#             first_value[state] = np.array([init[state]]*n)
    
#     init_array = []
#     for state in states:
#         if len(first_value[state]) != n:
#             raise Exception('Error in ' + state + '.\nMake sure all data has the same number of scenarios.')
#         init_array.append(first_value[state])
    
#     init_array = np.array(init_array).T
#     init_dict  = {i+1: init_array[i] for i in range(len(init_array))}
#     return init_dict

# def get_states(states, data_dict):
#     '''
#     Supporting function for extrac_from_timeseries. Do not run.
#     '''
#     return {state: data_dict[state] for state in states if data_dict.get(state)}

# def make_tspan(data_dict):
#     '''
#     Supporting function for extrac_from_timeseries. Do not run.
#     '''
#     tspan = np.concatenate([data_dict[state][0] for state in data_dict])
#     return np.unique(tspan)
    



import warnings

# try to import tesbml or libsbml
# if both of these fail, libsbml cannot be imported - cannot continue
try:
    import tesbml as libsbml   
except ImportError:
    import libsbml

from math import isnan
from re import sub
import os, sys

def _isSBMLModel(obj):
  """
  Tests if object is a libsbml model
  """
  cls_stg = str(type(obj))
  if ('Model' in cls_stg) and ('lib' in cls_stg):
    return True
  else:
    return False

def _checkSBMLDocument(document): 
  document.getErrorLog().getNumFailsWithSeverity(2)
  if (document.getNumErrors() > 0):
    raise ValueError("Errors in SBML document")


# Some convenience methods    
def loadSBMLStr (sbmlStr):
    """
    Load an SBML model in the form of a string and return an instance to SBMLModel
    """
    return SbmlModel (sbmlStr=sbmlStr)

def loadSBMLFile (sbmlFile):
    """
    Load an SBML model from a file and return an instance to SBMLModel
    """
    return SbmlModel (sbmlFile=sbmlFile)

class SbmlModel(object):

    """   
    SbmlModel is used to construct simple models using libSBML methods and
    print out the model in SBML format.  A user can add species, parameters,
    reactions, events, assignment rules, rate rules, and initial assignments
    to a model.  Then, the user can view the model in SBML format by printing
    the string representation of the class.

    SbmlModel contains two attributes: *document*, an
    `SBMLDocument <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_s_b_m_l_document.html>`_
    object, an
    *model*, the
    `Model <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_model.html>`_
    attribute of *document*.
    
    You can also pass a SBML string and use simlesbml to obtain information about the model
    using the get methods, e.g
       
    model = simplesbml.SbmlModel (sbmlStr=mySBMLString)
    """

    def __init__(self, time_units='second', extent_units='mole', \
                 sub_units='mole', level=3, version=1, sbmlStr=None, sbmlFile=None):
        if level == 1:
            raise SystemExit('Error: SimpleSBML does not support SBML level 1')
        try:
            self.document = libsbml.SBMLDocument(level,version)
        except ValueError:
            raise SystemExit('Could not create SBMLDocument object')
        
        if (sbmlFile != None) and (sbmlStr != None):
           raise Exception.Create ('Only select a string or file as the source of the SBML, not both')

        if (sbmlFile == None) and (sbmlStr == None): # They want to create a new model
            self.model = self.document.createModel()
            self._check(self.model, 'create model')
            if self.document.getLevel() == 3:
                self._check(self.model.setTimeUnits(time_units), 'set model-wide time units')
                self._check(self.model.setExtentUnits(extent_units), 'set model units of extent')
                self._check(self.model.setSubstanceUnits(sub_units),'set model substance units')

            per_second = self.model.createUnitDefinition()
            self._check(per_second,                         'create unit definition')
            self._check(per_second.setId('per_second'),     'set unit definition id')
            unit = per_second.createUnit()
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            
            
            per_min = self.model.createUnitDefinition()
            self._check(per_min,                          'create unit definition')
            self._check(per_min.setId('per_min'),      'set unit definition id')
            unit = per_min.createUnit()
            self._check(unit,                               'create unit on per_min')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(60),              'set unit multiplier')
            
            minute = self.model.createUnitDefinition()
            self._check(minute,                         'create unit definition')
            self._check(minute.setId('minute'),     'set unit definition id')
            unit = minute.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(60),              'set unit multiplier')

            seconds = self.model.createUnitDefinition()
            self._check(seconds,                         'create unit definition')
            self._check(seconds.setId('seconds'),     'set unit definition id')
            unit = seconds.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
        
            molperLmin = self.model.createUnitDefinition()
            self._check(molperLmin,                         'create unit definition')
            self._check(molperLmin.setId('molperLmin'),     'set unit definition id')
            unit = molperLmin.createUnit()
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_MOLE),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            unit = molperLmin.createUnit()
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_LITRE),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            unit = molperLmin.createUnit()
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(60),              'set unit multiplier')
            
            molpermin = self.model.createUnitDefinition()
            self._check(molpermin,                         'create unit definition')
            self._check(molpermin.setId('molpermin'),     'set unit definition id')
            unit = molpermin.createUnit()
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_MOLE),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            self._check(unit,                               'create unit on per_second')
            self._check(unit.setKind(libsbml.UNIT_KIND_SECOND),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(60),              'set unit multiplier')
            
            molperL = self.model.createUnitDefinition()
            self._check(molperL,                         'create unit definition')
            self._check(molperL.setId('molperL'),     'set unit definition id')
            unit = molperL.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_MOLE),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            unit = molperL.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_LITRE),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            
            mol = self.model.createUnitDefinition()
            self._check(mol,                         'create unit definition')
            self._check(mol.setId('mol'),     'set unit definition id')
            unit = mol.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_MOLE),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            
            Dimension_less = self.model.createUnitDefinition()
            self._check(Dimension_less,                         'create unit definition')
            self._check(Dimension_less.setId('Dimension_less'),     'set unit definition id')
            unit = Dimension_less.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_DIMENSIONLESS),     'set unit kind')
            self._check(unit.setExponent(0),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(0),              'set unit multiplier')
            
            items = self.model.createUnitDefinition()
            self._check(items,                         'create unit definition')
            self._check(items.setId('items'),     'set unit definition id')
            unit = items.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_ITEM),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            
            itemspermol = self.model.createUnitDefinition()
            self._check(itemspermol,                         'create unit definition')
            self._check(itemspermol.setId('itemspermol'),     'set unit definition id')
            unit = itemspermol.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_ITEM),     'set unit kind')
            self._check(unit.setExponent(1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')
            unit = itemspermol.createUnit()
            self._check(unit,                               'create unit')
            self._check(unit.setKind(libsbml.UNIT_KIND_MOLE),     'set unit kind')
            self._check(unit.setExponent(-1),               'set unit exponent')
            self._check(unit.setScale(0),                   'set unit scale')
            self._check(unit.setMultiplier(1),              'set unit multiplier')

            self.addCompartment()
        else:
            if sbmlStr != None:
                reader = libsbml.SBMLReader()  
                self.document = reader.readSBMLFromString(sbmlStr)

            if sbmlFile != None:
                if os.path.isfile(sbmlFile):
                   with open(sbmlFile, 'r') as fd:
                        xml = fd.read()
                   reader = libsbml.SBMLReader()  
                   self.document = reader.readSBMLFromString(xml)
                else:
                   raise Exception ('Specified file name does not appear to be a file?')    

            _checkSBMLDocument(self.document)
            self.model = self.document.getModel()
    
    def _check(self, value, message):
        if value == None:
            raise SystemExit('LibSBML returned a null value trying to ' + \
                    message + '.')
        elif type(value) is int:
            if value == libsbml.LIBSBML_OPERATION_SUCCESS:
                return
            else:
                err_msg = 'Error trying to ' + message + '.' \
                + 'LibSBML returned error code ' + str(value) + ': "'\
                + libsbml.OperationReturnValue_toString(value).strip() + '"'
            raise RuntimeError(err_msg)
        else:
            return

    def addCompartment(self, vol=1, comp_id=''):
        """Adds a `Compartment <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_compartment.html>`_
        of volume *vol* litres to the model.  The default volume is 1 litre. If the user does
        not specify *comp_id*, the id is set to 'c<n>' where the new compartment
        is the nth compartment added to the model. All SbmlModel objects are
        initialized with a default compartment 'c1'."""

        c1 = self.model.createCompartment()
        self._check(c1,                                 'create compartment')
        if len(comp_id) == 0:
            comp_id = 'c' + str(self.model.getNumCompartments())
        self._check(c1.setId(comp_id),                     'set compartment id')
        self._check(c1.setConstant(True),               'set compartment "constant"')
        self._check(c1.setSpatialDimensions(3),         'set compartment dimensions')

        self._check(c1.setSize(vol),                      'set compartment "size"')
        self._check(c1.setUnits('litre'),               'set compartment size units')
        return c1

    def addSpecies(self, species_id, amt, comp='c1'):
        
        """Adds a `Species <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_species.html>`_
        to the model.  If *species_id* starts
        with the '$' symbol, the species is set as a boundary condition.  If
        *species_id* is enclosed in brackets, the *amt* is the initial
        concentration of species within the specified compartment in mol/L.
        Otherwise, *amt* is the initial amount of species in moles."""

        s1 = self.model.createSpecies()
        self._check(s1,                           'create species s1')
        self._check(s1.setCompartment(comp),      'set species s1 compartment')
        if species_id[0] == '[' and species_id[len(species_id)-1] == ']':
            self._check(s1.setInitialConcentration(amt),    'set initial concentration for s1')
            species_id = species_id[1:(len(species_id)-1)]
        else:
            self._check(s1.setInitialAmount(amt),     'set initial amount for s1')
        self._check(s1.setSubstanceUnits(self.model.getSubstanceUnits()), 'set substance units for s1')
        if species_id[0] == '$':
            self._check(s1.setBoundaryCondition(True), \
                    'set "boundaryCondition" on s1')
            self._check(s1.setConstant(False), 'set "constant" attribute on s1')
            self._check(s1.setId(species_id[1:len(species_id)]), 'set species s1 id')
        else:
            self._check(s1.setBoundaryCondition(False), \
                    'set "boundaryCondition" on s1')
            self._check(s1.setConstant(False), 'set "constant" attribute on s1')
            self._check(s1.setId(species_id),  'set species s1 id')
        self._check(s1.setHasOnlySubstanceUnits(False), \
                'set "hasOnlySubstanceUnits" on s1')
        return s1

    def addParameter(self, param_id, val, units='per_second'):      
        """Adds a `Parameter <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_parameter.html>`_
        to the model.  *val* is the value of the parameter, and
        *param_id* is the parameter id.  If units are not specified by the user,
        the default units are 1/sec."""

        k = self.model.createParameter()
        self._check(k,                        'create parameter k')
        self._check(k.setId(param_id),        'set parameter k id')
        self._check(k.setConstant(False),      'set parameter k "not constant"')
        self._check(k.setValue(val),          'set parameter k value')
        self._check(k.setUnits(units), 'set parameter k units')
        return k

    def addReaction(self, reactants, products, expression, local_params={}, rxn_id=''):      
        """Adds a `Reaction <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_reaction.html>`_
        to the model.

        *reactants* and *products* are lists of species
        ids that the user wishes to define as reactants and products, respectively.
        If one of these lists contains a string with a number followed by a species
        id (i.e. '2 G6P') then the number is interpreted as the stoichiometry of
        the species.  Otherwise it is assumed to be 1.

        *expression* is a string that represents the reaction rate expression.

        *local_params* is a dictionary where the keys are local parameter ids and
        the values are the desired values of the respective parameters.

        If the
        user does not define a reaction id, it is set as 'v<n>' where the new
        reaction is the nth reaction added."""

        r1 = self.model.createReaction()
        self._check(r1,                         'create reaction')
        if len(rxn_id) == 0:
            rxn_id = 'v' + str(self.model.getNumReactions())
        self._check(r1.setId(rxn_id),           'set reaction id')
        self._check(r1.setReversible(False),    'set reaction reversibility flag')
        self._check(r1.setFast(False),          'set reaction "fast" attribute')

        for re in reactants:
            if re is not None and '$' in re:
                re.translate(None, '$')
            re_split = re.split()
            if len(re_split) == 1:
                sto = 1.0
                re_id = re
            elif len(re_split) == 2 and re_split[0].isdigit():
                sto = float(re_split[0])
                re_id = re_split[1]
            else:
                err_msg = 'Error: reactants must be listed in format \'S\' or \'(float)\' S\''
                raise SystemExit(err_msg)
            s1 = self.model.getSpecies(re_id)
            species_ref1 = r1.createReactant()
            self._check(species_ref1,                       'create reactant')
            self._check(species_ref1.setSpecies(s1.getId()), \
                    'assign reactant species')
            self._check(species_ref1.setStoichiometry(sto), \
                    'assign reactant stoichiometry')
            if self.document.getLevel() == 3:
                self._check(species_ref1.setConstant(True), \
                    'set "constant" on species ref 1')

        for pro in products:
            if pro is not None and '$' in pro:
                pro.translate(None, '$')
            pro_split = pro.split()
            if len(pro_split) == 1:
                sto = 1.0
                pro_id = pro
            elif len(pro_split) == 2:
                sto = float(pro_split[0])
                pro_id = pro_split[1]
            else:
                err_msg = 'Error: products must be listed in format \'S\' or \'(float)\' S\''
                raise SystemExit(err_msg)
            s2 = self.model.getSpecies(pro_id)
            species_ref2 = r1.createProduct()
            self._check(species_ref2, 'create product')
            self._check(species_ref2.setSpecies(s2.getId()), \
                    'assign product species')
            self._check(species_ref2.setStoichiometry(sto), \
                    'set product stoichiometry')
            if self.document.getLevel() == 3:
                self._check(species_ref2.setConstant(True), \
                    'set "constant" on species ref 2')

        math_ast = libsbml.parseL3Formula(expression)
        self._check(math_ast,    'create AST for rate expression')

        kinetic_law = r1.createKineticLaw()
        self._check(kinetic_law,                   'create kinetic law')
        self._check(kinetic_law.setMath(math_ast), 'set math on kinetic law')
        for param in local_params.keys():
            val = local_params.get(param)
            if self.document.getLevel() == 3:
                p = kinetic_law.createLocalParameter()
            else:
                p = kinetic_law.createParameter()
            self._check(p, 'create local parameter')
            self._check(p.setId(param), 'set id of local parameter')
            self._check(p.setValue(val),   'set value of local parameter')
        return r1

    def addEvent(self, trigger, assignments, persistent=True, \
                 initial_value=False, priority=0, delay=0, event_id=''):           
        """Adds an `Event <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_event.html>`_
        to the model.

        *trigger* is the string representation of a
        logical expression that defines when an event is 'triggered', meaning
        when the event is ready to be executed.

        *delay* is a numerical value that defines the amount of time between when
        the event is triggered and when the event assignment is implemented, in
        previously defined model-wide time units.

        *assignments* is a dictionary
        where the keys are variables to be changed and the values are the
        variables' new values.

        *persistent* is a boolean that defines whether the
        event will still be executed if the trigger switches from ``True`` to ``False``
        between the event's trigger and its execution.

        *initial_value* is the value of *trigger* when t < 0.

        *priority* is a numerical value that determines
        which event is executed if two events are executed at the same time.  The
        event with the larger ``priority`` is executed.

        .. note:: An event is only triggered when the trigger switches from ``False`` to
            ``True``.  If the trigger's initial value is ``True``, the event will not be
            triggered until the value switches to ``False`` and then back to ``True``."""

        e1 = self.model.createEvent()
        self._check(e1,     'create event')
        if len(event_id) == 0:
            event_id = 'e' + str(self.model.getNumEvents())
        self._check(e1.setId(event_id),    'add id to event')
        if self.document.getLevel()==3 or (self.document.getLevel()==2 \
                    and self.document.getVersion()==4):
            self._check(e1.setUseValuesFromTriggerTime(True), 'set use values from trigger time')

        tri = e1.createTrigger()
        self._check(tri,  'add trigger to event')
        tri_ast = libsbml.parseL3Formula(trigger)
        self._check(tri.setMath(tri_ast),     'add formula to trigger')
        if self.document.getLevel() == 3:
            self._check(tri.setPersistent(persistent),   'set persistence of trigger')
            self._check(tri.setInitialValue(initial_value), 'set initial value of trigger')

        de = e1.createDelay()
        if self.document.getLevel() == 3:
            k = self.addParameter(event_id+'Delay', delay, self.model.getTimeUnits())
        else:
            k = self.addParameter(event_id+'Delay', delay, 'time')
        self._check(de,               'add delay to event')
        delay_ast = libsbml.parseL3Formula(k.getId())
        self._check(de.setMath(delay_ast),     'set formula for delay')

        for a in assignments.keys():
            assign = e1.createEventAssignment()
            self._check(assign,   'add event assignment to event')
            self._check(assign.setVariable(a),  'add variable to event assignment')
            val_ast = libsbml.parseL3Formula(assignments.get(a))
            self._check(assign.setMath(val_ast),    'add value to event assignment')

        if self.document.getLevel() == 3:
            pri = e1.createPriority()
            pri_ast = libsbml.parseL3Formula(str(priority))
            self._check(pri.setMath(pri_ast), 'add priority to event')
        return e1

    def addAssignmentRule(self, var, math):
        
        """Adds an `AssignmentRule <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_assignment_rule.html>`_
        to the model.  An assignment rule is an equation
        where one side is equal to the value of a state variable and the other side
        is equal to some expression.  *var* is the id of the state variable and *math*
        is the string representation of the expression."""

        r = self.model.createAssignmentRule()
        self._check(r,                        'create assignment rule r')
        self._check(r.setVariable(var),          'set assignment rule variable')
        math_ast = libsbml.parseL3Formula(math)
        self._check(r.setMath(math_ast), 'set assignment rule equation')
        return r

    def addRateRule(self, var, math):
 
        """Adds a `RateRule <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_rate_rule.html>`_
        to the model.  A rate rule is similar to an assignment
        rule, but instead of describing a state variable's value as an
        expression, it describes the derivative of the state variable's value
        with respect to time as an expression.  *var* is the id of the state variable
        and *math* is the string representation of the expression."""

        r = self.model.createRateRule()
        self._check(r,                        'create rate rule r')
        self._check(r.setVariable(var),          'set rate rule variable')
        math_ast = libsbml.parseL3Formula(math)
        self._check(r.setMath(math_ast), 'set rate rule equation')
        return r

    def addInitialAssignment(self, symbol, math):
        
        """Adds an `InitialAssignment <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_initial_assignment.html>`_
        to the model.  If the
        initial value of a variable depends on other variables or parameters, this
        method can be used to define an expression that describes the initial value
        of the variable in terms of other variables or parameters.  *symbol* is the
        id of the variable and *math* is the string representation of the expression."""

        if self.document.getLevel() == 2 and self.document.getVersion() == 1:
            raise SystemExit('Error: InitialAssignment does not exist for \
                    this level and version.')
        a = self.model.createInitialAssignment()
        self._check(a,   'create initial assignment a')
        self._check(a.setSymbol(symbol),    'set initial assignment a symbol')
        math_ast = libsbml.parseL3Formula(math)
        self._check(a.setMath(math_ast),    'set initial assignment a math')
        return a

    def setLevelAndVersion(self, level, version):
        if level == 2 and version == 1:
            self._check(self.document.checkL2v1Compatibility(), 'convert to level 2 version 1')
        elif level == 2 and version == 2:
            self._check(self.document.checkL2v2Compatibility(), 'convert to level 2 version 2')
        elif level == 2 and version == 3:
            self._check(self.document.checkL2v3Compatibility(), 'convert to level 2 version 3')
        elif level == 2 and version == 4:
            self._check(self.document.checkL2v4Compatibility(), 'convert to level 2 version 4')
        elif level == 3 and version == 1:
            self._check(self.document.checkL3v1Compatibility(), 'convert to level 3 version 1')
        else:
            raise SystemExit('Invalid level/version combination')

        isSet = self.document.setLevelAndVersion(level, version)
        self._check(isSet, 'convert to level ' + str(level) + ' version ' + str(version))

    def getDocument(self):
        
        """Returns the
        `SBMLDocument <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_s_b_m_l_document.html>`_
        object of the sbmlModel. 

        This is not something you need to care about unless you need direct access to libsbml"""
        return self.document

    def getModel(self):
        """Returns the
        `Model <http://sbml.org/Software/libSBML/docs/python-api/classlibsbml_1_1_model.html>`_
        object of the sbmlModel.

        This is not something you need to care about unless you need direct access to libsbml"""
        return self.model

    def getModelId (self):
        """
        Returns the SBML Id given to the model. 
        """
        return self.model.getModel().getId()
    
    def getNumCompartments (self):
        """
        Returns the number of compartments in the current model.
        """
        return self.model.getNumCompartments()   
       
    def getNumSpecies (self):
        """
        Returns the number of all species in the current model.
        """
        return self.model.getNumSpecies()

    def getNumParameters (self):
        """
        Returns the number of global parameters in the current model.
        """        
        return self.model.getNumParameters()
    
    def getNumReactions (self):
        """
        Returns the number of reactions in the current model.
        """        
        return self.model.getNumReactions()

    def getNumEvents (self):
        """
        Returns the number of events in the current model.
        """        
        return self.model.getNumEvents()    

    def getNumRules (self):
        """
        Returns the number of rules in the current model.
        """        
        return self.model.getNumRules()  

    def getNumFunctionDefinitions (self):
        """
        Returns the number of function definitions in the current model.
        """        
        return self.model.getNumFunctionDefinitions()
    
    def getNumInitialAssignments (self):
        """
        Returns the number of initial assignments in the current model.
        """        
        return self.model.getNumInitialAssignments() 
    
    #def getSpecies(self, species_id):      
    #    return self.model.getSpecies(species_id)

    def getListOfCompartmentIds(self):
        """
        Returns a list of compartment Ids
        """
        alist = []
        nCompartments = self.model.getNumCompartments()
        for i in range (nCompartments):
           comp = self.model.getCompartment(i)
           alist.append (comp.getId())
        return alist
    

    def getCompartmentId (self, index):
        """
        Returns the indexth compartment from the list of compartments
        """
        return self.getListOfCompartmentIds()[index]

    def getCompartmentVolume (self, Id):
        """
        **Parameters**
        
           Id: The Id of the compartment in question
           
        **Returns**
        
           The volume of the specified compartment
        """
        p = self.model.getCompartment(Id)
        if p != None:
           return p.getVolume()
        raise Exception ('Compartment does not exist')         
        
    # --------------------------------------------------------------
    def getListOfAllSpecies(self):
        """
        Returns a list of **ALL** species Ids in the model
        """
        alist = []
        nSpecies = self.model.getNumSpecies()
        for i in range (nSpecies):
           sp = self.model.getSpecies(i)
           alist.append (sp.getId())         
        return alist
    
    def getNthFloatingSpeciesId (self, index):
        """
        Returns the Id of the nth floating species
        """
        return self.getListOfFloatingSpecies()[index]

    def getNthBoundarySpeciesId (self, index):
        """
        Returns the Id of the nth boundary species
        """
        return self.getListOfBoundarySpecies()[index]

    def getCompartmentIdSpeciesIsIn (self, speciesId):
        """
        Returns the compartment Id that the species givenin th argument is in.
        """
        sp = self.model.getSpecies(speciesId)
        if sp != None:
            return sp.getCompartment()
        else:
            raise Exception ('Species Id:' + speciesId + ' does not exist in the model') 

    def isSpeciesValueSet (self, Id):
        """
        Returns true if the species has been set a value (concentration or amount).

        Example: if model.isSpeciesValueSet ('ATP'):
        """
        p = self.model.getSpecies(Id)
        if p == None:
           raise Exception ('Species does not exist')  
        if p.isSetInitialConcentration() or p.isSetInitialAmount():
            return True
        else:
            return False

    def getSpeciesInitialConcentration (self, Id):
        """
        Returns the initial values for the concentration of a species with given Id

        The species can be an index to the indexth species or the Id of the species. You can get the Ids by calling getListOfSpecies()

        Example: value = model.getSpeciesInitialConcentration ('Glucose')
        """
        p = self.model.getSpecies(Id)
        if p == None:
           raise Exception ('Species does not exist')             
        if p.isSetInitialConcentration() == True:
           return p.getInitialConcentration()
        else:
           raise Exception ('Species has not been initialized with a concentration')
        

    def getSpeciesInitialAmount (self, Id):
        """
        Returns the initial values for the amount of a species with given Id

        The species can be an index to the indexth species or the Id of the species. You can get the Ids by calling getListOfSpecies()

        Example: value = model.getSpeciesInitialAmount ('Glucose')
        """
        p = self.model.getSpecies(Id)
        if p == None:
           raise Exception ('Species does not exist')   
        if p.isSetInitialAmount():
           return p.getInitialAmount()
        else:
           raise Exception ('Species has not been initialized with an amount') 
         
    def getListOfFloatingSpecies(self):
        """
        Returns a list of all floating species Ids.
        """
        alist = []
        nSpecies = self.model.getNumSpecies()
        for i in range (nSpecies):
            sp = self.model.getSpecies(i)
            if not sp.getBoundaryCondition():
               alist.append (sp.getId())         
        return alist    

    def getListOfBoundarySpecies(self):
        """
        Returns a list of all boundary species Ids.
        """
        alist = []
        nSpecies = self.model.getNumSpecies()
        for i in range (nSpecies):
            sp = self.model.getSpecies(i)            
            if sp.getBoundaryCondition():
               alist.append (sp.getId())         
        return alist  
    
    def isFloatingSpecies (self, Id):
        """
        Returns true if the Id is a floating species
        """
        sp = self.model.getSpecies(Id)
        if sp != None:
           if sp.getBoundaryCondition():
              return False
           else:
              return True
        else:
           raise Exception ('Id: ' + Id + ' in isFloatingSpecies is not a species Id')  

    def isBoundarySpecies (self, Id):
        """
        Returns true if the Id is a boundary species
        """
        sp = self.model.getSpecies(Id)
        if sp != None:
           if sp.getBoundaryCondition():
              return True
           else:
              return False
        else:
           raise Exception ('Id: ' + Id + ' in isFloatingSpecies is not a species Id')  

    def isAmount (self, Id):
        """
        Returns true if species Id has assigned to it an amount rather then a concentration"
        """
        sp = self.model.getSpecies(Id)
        if sp != None:
            if sp.isSetInitialAmount():
                return True
            else:
                return False
        else:
           raise Exception ('Id: ' + Id + ' in isAmount is not a species Id')  

    def isConcentration (self, Id):
        """
        Returns true if species Id has assigned to it a concentration rather then an amount"
        """
        sp = self.model.getSpecies(Id)
        if sp != None:
            if sp.getInitialConcentration():
                return True
            else:
                return False
        else:
           raise Exception ('Id: ' + Id + ' in isConcentration is not a species Id')  


    def getNumFloatingSpecies (self):
        """
        Returns the number of floating species.
        """
        return len (self.getListOfFloatingSpecies())

    def getNumBoundarySpecies (self):
        """
        Returns the number of boundary species.
        """
        return len (self.getListOfBoundarySpecies())
    
    def getListOfParameterIds(self):
        """
        Returns a list of all global parameter Ids in the model.
        """
        alist = []
        nParameters = self.model.getNumParameters() 
        for i in range (nParameters):
            p = self.model.getParameter(i)
            alist.append (p.getId())             
        return alist

    def getParameterId (self, index):
        """
        Returns the indexth parameter from the list of parameter
        """
        return self.getListOfParameterIds()[index]
    
    def isParameterValueSet (self, Id):
        """
        Returns true if the parameter has been assigned a value

        Example: if model.isParameterValueSet ('k1'):

        """
        p = self.model.getParameter(Id)
        if p == None:
           raise Exception ('Parameter does not exist')  
        if p.isSetValue():
            return True
        else:
            return False

    def getParameterValue (self, Id):
        """
        Returns the value for a given model parameter.

        **Parameters**
        
           Id: the Id of the parameter in question

        Example: value = model.getParameterValue ('k1')       
        """      
        p = self.model.getParameter(Id)
        if p != None:
           if p.isSetValue():
              return p.getValue()
           else:
              return 0
        else:
           raise Exception ('Parameter does not exist')      
        
    def getListOfReactionIds (self):
        """
        Returns a list of all reaction Ids.
        """       
        alist = []
        nReactions = self.model.getNumReactions() 
        for i in range (nReactions):
            p = self.model.getReaction(i)
            alist.append (p.getId())             
        return alist
    
    def getNthReactionId (self, index):
        """
        Returns the Id of the nth reaction
        """
        return self.getListOfReactionIds()[index]

    def getNumReactants (self, Id):
        """
         Returns the number of reactants in the reaction given by the Id argument.

        **Parameters**
        
           Id (string): The Id of the reaction in question.

        Example: numProducts = model.getNumReactants ('J1')
        """ 
        p = self.model.getReaction(Id)
        if p != None:
           return p.getNumReactants()
        raise Exception ('Reaction does not exist')  
        
    def getNumProducts (self, Id):
        """
        Returns the number of products in the reaction given by the Id argument.

        **Parameters**
        
           Id (string): The Id of the reaction in question.
           
        Example: numProducts = model.getNumProducts ('J1')
        """ 
        p = self.model.getReaction(Id)
        if p != None:
           return p.getNumProducts()
        raise Exception ('Reaction does not exist') 
        
    def getRateLaw (self, Id):
        """
        Returns the expression for the rate laws of the reaction given by the specified Id

        **Parameters**
        
           Id (string): The Id of the reaction in question.
  
        Example: formulaStr = model.getRateLaw ('J1')
        """ 
        p = self.model.getReaction(Id)
        if p != None:
           return p.getKineticLaw().getFormula()
        raise Exception ('Reaction does not exist')      
              
    def getReactant (self, reactionId, reactantIndex):
        """
        Returns the Id of the reactantIndexth reactant in the reaction given by the reactionId

        **Parameters**
        
           Id (string): The Id of the reaction in question
           
           reactantIndex (int): The ith reactant in the reaction

        Example: astr = model.getReactant ('J1', 0)
        """ 
        ra = self.model.getReaction(reactionId)
        sr = ra.getReactant(reactantIndex)
        return sr.getSpecies()
    
    def getProduct (self, reactionId, productIndex):
        """
        Returns the Id of the productIndexth product in the reaction given by the reactionId

        **Parameters**
        
           Id (string): The Id of the reaction in question
           
           productIndex (int): The ith product in the reaction

        Example: astr = model.getProduct ('J1', 0)
        """ 
        ra = self.model.getReaction(reactionId)
        sr = ra.getProduct(productIndex)
        return sr.getSpecies()
    
    def getReactantStoichiometry (self, reactionId, reactantIndex):
        """
        Returns the stoichiometry for a reactant in a reaction.

        **Parameters**
        
           reactionId (string): The Id of the reaction in question
           
           reactantIndex (int): The ith reactant in the reaction
           
        **Returns**
        
           The value of the reactant stoichiometry

        Example: stInt = model.getReactantStoichiometry ('J1', 0)
        """         
        ra = self.model.getReaction(reactionId)
        if ra != None:
           sr = ra.getReactant(reactantIndex)
           return sr.getStoichiometry ()
        else:
           raise Exception ('Reaction:' + reactionId + ' does not exist') 

    def getProductStoichiometry (self, reactionId, productIndex):
        """
        Returns the stoichiometry for a product in a reaction.

        **Parameters**
        
           reactionId (string): The Id of the reaction in question
           
           productIndex (int): The ith product in the reaction
           
        **Returns**
        
           The value of the product stoichiometry

        Example: stInt = model.getProductStoichiometry ('J1', 0)
        """
        ra = self.model.getReaction(reactionId)
        if ra != None:
           sr = ra.getProduct(productIndex)
           return sr.getStoichiometry ()
        else:
           raise Exception ('Reaction:' + reactionId + ' does not exist') 
        
    def getNumModifiers (self, reactionId):
        """
        Returns the number of modifiers in a given reaction.
        """
        return self.model.getReaction (reactionId).getNumModifiers()

    def getListOfModifiers (self, reactionId):
        """
        Returns the list of modifiers in a given reaction.
        """
        alist = []
        reaction = self.model.getReaction (reactionId)
        if reaction != None:
           for i in range (reaction.getNumModifiers()):
               alist.append (reaction.getModifier (0).getSpecies())
           return alist
        else:
           raise Exception ('No such reaction can be found:', reactionId)
 
    def getListOfRuleIds(self):
        """
        Returns a list of Ids for the rules
        """ 
        alist = []
        nRules = self.model.getNumRules() 
        for i in range (nRules):
            p = self.model.getRule(i)
            alist.append (p.getId())             
        return alist
    
    def getRuleId (self, index):
        """
        Returns the rule Id of the indexth rule

        Example: rule = model.getRuleId (2)
        """         
        rule = self.model.getRule (index)
        if rule != None:
            return rule.getId()

    def getRuleRightSide (self, rule):
        """
        Returns the formula on the right-hand side of the rule in question.

        The rule can be an index to the indexth rule or the Id of the rule. You can get the Ids by calling getListOfRules()

        Example: formula = model.getRuleRightSide (0)
        """
        rule = self.model.getRule (rule)
        if rule != None:
            return rule.getFormula()

    def getRuleType (self, rule):
        """
        Returns a string indicating the type of rule in question.

        The rule can be an index to the indexth rule or the Id of the rule. You can get the Ids by calling getListOfRules()

        Example: ruleStr = model.getRuleType (0)
        """
        myRule = self.model.getRule (rule)
        if myRule != None:
           t1 = myRule.getTypeCode()
           if t1 == libsbml.SBML_RATE_RULE:
              return 'ODE (or rate) rule' 
           if t1 == libsbml.SBML_ASSIGNMENT_RULE:
              return 'Assignment rule' 
           if t1 == libsbml.SBML_ALGEBRAIC_RULE:
              return 'Algebraic rule' 
           raise Exception ('Unknown rule in SBML model')

    def isRuleType_Assignment (self, rule):
        """
        Returns true the rule is an assignemnt type
        """
        myRule = self.model.getRule (rule)
        if myRule != None:
           t1 = myRule.getTypeCode()
           if t1 == libsbml.SBML_ASSIGNMENT_RULE:
              return True
           else:
              return False 
        else:
           return None
           #raise Exception ('The rule: ' + str (rule) + ', does not exist') 

    def isRuleType_Rate (self, rule):
        """
        Returns true the rule is an rate rule (ode) type
        """
        myRule = self.model.getRule (rule)
        if myRule != None:
           t1 = myRule.getTypeCode()
           if t1 == libsbml.SBML_RATE_RULE:
              return True
           else:
              return False 
        else:
           raise Exception ('The rule: ' + str (rule) + ', does not exist') 

    def isRuleType_Algebraic (self, rule):
        """
        Returns true the rule is an algebraic rule type
        """
        myRule = self.model.getRule (rule)
        if myRule != None:
           t1 = myRule.getTypeCode()
           if t1 == libsbml.SBML_ALGEBRAIC_RULE:
              return True
           else:
              return False 
        else:
           raise Exception ('The rule: ' + str (rule) + ', does not exist') 

    #  ----------------------- EVENTS -------------------------
    def getEventId (self, index):
        """
        Returns the Id for the indexth event

        Example: astr = model.getEventId (0)
        """
        return self.model.getEvent (index).getId()

    def getEventString (self, event):
        """
        Returns the indexth event as a compelte string.

        The event argument can be an index to the indexth event or the Id of the event  

        Example: print (model.getEventString (0))
        """
        myEvent = self.model.getEvent (event)
        astr = 'at ' + self.getEventTrigger (event) + ' then { '
        num = myEvent.getNumEventAssignments()
        astr = astr + self.getEventVariable (event, 0) + ' = ' + self.getEventAssignment (event, 0)
        for i in range (1, num):
            astr = astr + '; ' + self.getEventVariable (event, i) + ' = ' + self.getEventAssignment (event, i)
        astr = astr + ' }'
        return astr
        
    def getEventTrigger (self, event):
        """
        Returns the formula as a string for the event trigger of the event event.

        The event argument can be an index to the indexth event or the Id of the event

        Example : astr = model.getEventTrigger (0)
        """
        myEvent = self.model.getEvent(event)
        trig = myEvent.getTrigger()        
        return libsbml.formulaToL3String (trig.getMath())

    def getNumEventAssignments (self, index):
        """
        Returns the number of assignments in the indexth rule.
        """
        event = self.model.getEvent(index)
        return event.getNumEventAssignments()

    def getEventVariable (self, event, assignmentIndex):
        """
        Returns the event variables (i.e the left-hand side) for the assignmentIndexth assignment
        in the event, given by event,

        The event argument can be an index to the indexth event or the Id of the event

        Example: astr = model.getEventVariable (0, 0)
        """
        myEvent = self.model.getEvent(event)
        eventAss = myEvent.getEventAssignment(assignmentIndex)  
        return eventAss.getVariable()

    def getEventAssignment (self, event, assignmentIndex):
        """
        Retuns the assignmentIndexth assignment in the event event.

        The event argument can be an index to the indexth event or the Id of the event   

        Example: mathStr = model.getEventAssignment (1, 0)     
        """
        myEvent = self.model.getEvent(event)
        eventAss = myEvent.getEventAssignment(assignmentIndex)                
        m = eventAss.getMath()
        return libsbml.formulaToL3String(m)

    def getListOfFunctionIds (self):
        """
        Returns a list of function definition Ids
        """
        alist = []
        for i in range (self.model.getNumFunctionDefinitions()):
            alist.append (self.model.getFunctionDefinition (i).getId())
        return alist

    def getNumFunctionDefinitions(self):
        """
        Returns the number of user defined functions in the model
        """
        return self.model.getNumFunctionDefinitions()

    def getFunctionId (self, index):
        """
        Retuns the Id of the indexth user function definition

        Example: mathStr = model.getFunctionId (0)     
        """
        return self.model.getFunctionDefinition (index).getId()

    def getFunctionBody (self, func):
        """
        """ 
        myFunc = self.model.getFunctionDefinition (func) 
        return libsbml.formulaToL3String(myFunc.getBody())

    def getNumArgumentsInUserFunction (self, func):
        """
        Returns the number of arguments in the given function. The func argument can either be the name
        of a user function or its index.

        Example: 

          nargs = model.getNumArgumentsInUserFunction ('Hill')

          nargs = model.getNumArgumentsInUserFunction (0)
        """
        myFunc = self.model.getFunctionDefinition (func) 
        return myFunc.getNumArguments()                

    def getListOfArgumentsInUserFunction (self, func):
        """
        Returns a list of arguments in the specified user function. The func argument can either be the name
        of a user function or its index in the list of user fucntions.

        Example: 

          alist = model.getListOfArgumentsInUserFunction ('Hill')

          alist = model.getListOfArgumentsInUserFunction (0)
        """
        myFunc = self.model.getFunctionDefinition (func) 
        alist = []
        for i in range (myFunc.getNumArguments()):
            alist.append (libsbml.formulaToL3String(myFunc.getArgument(i)))
        return alist

    def toSBML(self):
        """Returns the model in SBML format as a string.  Also checks model consistency
        and prints all errors and warnings.

        Example: print (model.toSBML())
        """
        self.document.setConsistencyChecks(libsbml.LIBSBML_CAT_UNITS_CONSISTENCY, False)
        errors = self.document.checkConsistency()
        if (errors > 0):
           for i in range(errors):
               print(self.document.getError(i).getSeverityAsString(), ": ", self.document.getError(i).getMessage()) 

        return libsbml.writeSBMLToString(self.document)

    def __repr__(self):
        return self.toSBML()

def writeCode(doc):
    """Returns a string containing calls to SimpleSBML functions that reproduce
    the model contained in the SBMLDocument *doc* in an sbmlModel object."""

    comp_template = 'model.addCompartment(vol=%s, comp_id=\'%s\');'
    species_template = 'model.addSpecies(species_id=\'%s\', amt=%s, comp=\'%s\');'
    param_template = 'model.addParameter(param_id=\'%s\', val=%s, units=\'%s\');'
    rxn_template = 'model.addReaction(reactants=%s, products=%s, expression=\'%s\', local_params=%s, rxn_id=\'%s\');'
    event_template = 'model.addEvent(trigger=\'%s\', assignments=%s, persistent=%s, initial_value=%s, priority=%s, delay=%s, event_id=\'%s\');'
    event_defaults = [True, False, '0', 0]
    assignrule_template = 'model.addAssignmentRule(var=\'%s\', math=\'%s\');'
    raterule_template = 'model.addRateRule(var=\'%s\', math=\'%s\');'
    initassign_template = 'model.addInitialAssignment(symbol=\'%s\', math=\'%s\')'
    init_template = 'import simplesbml\nmodel = simplesbml.sbmlModel(time_units=\'%s\', extent_units=\'%s\', sub_units=\'%s\', level=%s, version=%s);'
    init_defaults = ['second', 'mole', 'mole', 3, 1]
    command_list = []

    if doc.getLevel() == 1:
        warnings.warn('Warning: SimpleSBML does not support SBML Level 1.')

    props = libsbml.ConversionProperties()
    props.addOption('flatten comp', True)
    result = doc.convert(props)
    if(result != libsbml.LIBSBML_OPERATION_SUCCESS):
        raise SystemExit('Conversion failed: (' + str(result) + ')')

    mod = doc.getModel()
    comps = mod.getListOfCompartments()
    species = mod.getListOfSpecies()
    params = mod.getListOfParameters()
    rxns = mod.getListOfReactions()
    events = mod.getListOfEvents()
    rules = mod.getListOfRules()
    inits = []
    if doc.getLevel() == 3 or (doc.getLevel() == 2 and doc.getVersion() > 1):
        inits = mod.getListOfInitialAssignments()

    timeUnits = 'second'
    substanceUnits = 'mole'
    extentUnits = 'mole'
    if doc.getLevel() == 3:
        timeUnits = mod.getTimeUnits()
        extentUnits = mod.getExtentUnits()
        substanceUnits = mod.getSubstanceUnits()
    level = mod.getLevel()
    version = mod.getVersion()
    init_list = [timeUnits, extentUnits, substanceUnits, level, version]
    for i in range(0,5):
        if init_list[i] == init_defaults[i]:
            init_list[i] = 'del'

    command_list.append(init_template % \
            (init_list[0], init_list[1], init_list[2], init_list[3], init_list[4]))

    for comp in comps:
        if comp.getId() != 'c1':
            if comp.getId()[0] == 'c' and comp.getId()[1:len(comp.getId())].isdigit():
                if comp.getSize() == 1e-15:
                    command_list.append(comp_template % ('del', 'del'))
                else:
                    command_list.append(comp_template % (comp.getSize(), 'del'))
            else:
                if comp.getSize() == 1e-15:
                    command_list.append(comp_template % ('del', comp.getId()))
                else:
                    command_list.append(comp_template % (comp.getSize(), comp.getId()))

    for s in species:
        conc = s.getInitialConcentration()
        amt = s.getInitialAmount()
        sid = s.getId()
        if s.getCompartment() == 'c1':
            comp = 'del'
        else:
            comp = s.getCompartment()
        bc = s.getBoundaryCondition()
        if bc:
            sid = "$" + sid
        if isnan(conc) or amt > conc:
            command_list.append(species_template % (sid, str(amt), comp))
        else:
            command_list.append(species_template % ("[" + sid + "]", str(conc), comp))

    for p in params:
        val = p.getValue()
        pid = p.getId()
        if p.getUnits() == 'per_second':
            units = 'del'
        else:
            units = p.getUnits()
        isDelay = pid.find('Delay')
        if isDelay == -1:
            command_list.append(param_template % (pid, str(val), str(units)))

    for v in rxns:
        vid = v.getId()
        if vid[0] == 'v' and vid[1:len(vid)].isdigit():
            vid = 'del'
        reactants = []
        for r in v.getListOfReactants():
            reactants.append((str(r.getStoichiometry()) + ' ' + r.getSpecies()).replace('1.0 ', ''))
        products = []
        for p in v.getListOfProducts():
            products.append((str(p.getStoichiometry()) + ' ' + p.getSpecies()).replace('1.0 ', ''))
        expr = libsbml.formulaToString(v.getKineticLaw().getMath())
        local_params = {}
        local_ids = []
        local_values = []
        for k in v.getKineticLaw().getListOfParameters():
            local_ids.append(k.getId())
            local_values.append(k.getValue())
        local_params = dict(zip(local_ids, local_values))
        if len(local_params) == 0:
            local_params = 'del'
        command_list.append(rxn_template % (str(reactants), str(products), \
                    expr, str(local_params), vid))

    for e in events:
        persistent = True
        initialValue = False
        priority = '0'
        eid = e.getId()
        if len(eid) == 0 or (eid[0] == 'e' and eid[1:len(eid)].isdigit()):
            eid = 'del'
        if doc.getLevel() == 3:
            persistent = e.getTrigger().getPersistent()
            initialValue = e.getTrigger().getInitialValue()
            priority = e.getPriority()
            if type(priority) == libsbml.Priority:
                priority = libsbml.formulaToL3String(priority.getMath())
            else:
                priority = '0'
        tri = libsbml.formulaToL3String(e.getTrigger().getMath())
        did = e.getDelay()
        if type(did) == libsbml.Delay:
            delay = libsbml.formulaToL3String(did.getMath())
        else:
            delay = '0'
        assigns = e.getListOfEventAssignments()
        var = []
        values = []
        for assign in assigns:
            var.append(assign.getVariable())
            values.append(libsbml.formulaToL3String(assign.getMath()))
        assigns = dict(zip(var, values))

        event_list = [persistent, initialValue, priority, delay]
        for i in range(0,4):
            if event_list[i] == event_defaults[i]:
                event_list[i] = 'del'

        command_list.append(event_template % (tri, str(assigns), \
                event_list[0], event_list[1], event_list[2], event_list[3], eid))

    for r in rules:
        sym = r.getVariable()
        math = libsbml.formulaToL3String(r.getMath())
        if r.getTypeCode() == libsbml.SBML_ASSIGNMENT_RULE:
            command_list.append(assignrule_template % (sym, math))
        elif r.getTypeCode() == libsbml.SBML_RATE_RULE:
            command_list.append(raterule_template % (sym, math))
        else:
            next

    for i in inits:
        sym = i.getSymbol()
        math = libsbml.formulaToL3String(i.getMath())
        command_list.append(initassign_template % (sym, math))

    commands = '\n'.join(command_list)
    commands = sub('\w+=\'?del\'?(?=[,)])', '', commands)
    commands = sub('\((, )+', '(', commands)
    commands = sub('(, )+\)', ')', commands)
    commands = sub('(, )+', ', ', commands)
    return commands

def writeCodeFromFile(filename):
    """Reads the file saved under *filename* as an SBML format model and
    returns a string containing calls to SimpleSBML functions that reproduce
    the model in an sbmlModel object."""

    reader = libsbml.SBMLReader()
    doc = reader.readSBMLFromFile(filename)
    if doc.getNumErrors() > 0:
        raise SystemExit(doc.getError(0))
    return writeCode(doc)

def writeCodeFromString(sbmlstring):
    """Reads *sbmlstring* as an SBML format model and
    returns a string containing calls to SimpleSBML functions that reproduce
    the model in an sbmlModel object."""

    reader = libsbml.SBMLReader()
    doc = reader.readSBMLFromString(sbmlstring)
    return writeCode(doc)
